import asyncio
import websockets
import json
import os
from requests import Session
from requests.adapters import HTTPAdapter
from urllib3.util import Retry
import logging
from os.path import exists, expanduser
from getpass import getuser
from ndnsecure.auth import NdnSecureAuth
from .schema import RunModel
from urllib.parse import urlparse

log_level = os.getenv('ANSIBLE_LOG_LEVEL', 'WARN')
logging.basicConfig(level=getattr(logging, log_level))
logger = logging.getLogger('python-ansible-client')

PROD_API_URL = os.environ.get(
    'ANSIBLE_API_URL', 'https://ansible-service-master.dh-prod-eks.kubecluster.dreamhost.io/ansible')


class AnsibleClient():
    def __init__(self):

        self.token_file = os.path.expanduser('~%s/.ndnauth' % getuser())
        self.ansible_api_url = os.getenv('ANSIBLE_API_URL', PROD_API_URL)
        self.header = {
            "Content-Type": "application/json",
            'User-agent': 'python-ansible-client 1.0'}

        # setup auto-retry for 5XX status codes
        retry_strat = Retry(total=3, status_forcelist=list(
            range(500, 600)), backoff_factor=2, allowed_methods=frozenset(['GET', 'POST', 'PUT', 'DELETE']))
        adapter = HTTPAdapter(max_retries=retry_strat)

        self.httpsession = Session()
        self.httpsession.mount('http://', adapter)
        self.httpsession.mount('https://', adapter)

        # try to get API token from env var otherwise go through the login process
        self.api_token = os.getenv('ANSIBLE_API_TOKEN', None)
        self.api_token = self.api_token or self.fetch_authtoken()

        if self.api_token:
            self.header['Authorization'] = 'Bearer %s' % self.api_token

        self.refresh_uuid_token = None
        self.run_uuid_token = None  # celery uuid for scheduled ansible run
        self.host_refresh_uuid_token = None
        self.group_refresh_uuid_token = None
        # ID of ansible run (maps to service_queue_id, if provided)
        self.run_id = None
        self.inventory = None

    def api_auth(self):
        """
        Authenticate against the ansible api, fetching a token and storing it in ~/.ndnauth
        """
        logger.debug('Authenticating against ansible api login endpoint')
        header = self.header
        # remove auth header, as we're trying to authenticate now
        header.pop('Authorization', None)

        # try to get user/pass from ndnsecure/vault
        try:
            logger.debug("Attempting to load user/pass from NdnSecureAuth: %s" %
                         os.environ.get('ANSIBLE_API_SVC_ACCOUNT'))
            username, password = NdnSecureAuth(
                os.environ.get('ANSIBLE_API_SVC_ACCOUNT')).get()
        except TypeError:
            logger.debug(
                "Failed to load user/pass from NdnSecureAuth. Using environment variables")
            # if we got nothing, check env vars
            username = os.environ.get('ANSIBLE_API_USERNAME', None)
            password = os.environ.get('ANSIBLE_API_PASSWORD', None)

        # if we STILL got nothing, we cannot auth
        if not username or not password:
            logger.error(
                "Unable to load username/password for API authentication")
            raise Exception(
                "Unable to load username/password for API authentication")

        auth_resp = self.httpsession.post(
            ''.join([self.ansible_api_url, '/login']),
            headers=header,
            json={"username": username, "password": password})
        logger.debug('Received API response code: %s' % auth_resp.status_code)

        auth_resp.raise_for_status()
        if auth_resp:
            logger.debug("API authentication successful!")
            self.api_token = auth_resp.json()['access_token']
            self.header['Authorization'] = 'Bearer %s' % auth_resp.json()[
                'access_token']
            logger.debug("Saving token to %s" % self.token_file)
            with open(self.token_file, 'w+') as f:
                f.write(self.api_token)
            return auth_resp
        else:
            logger.debug("API authentication API failure: %s" %
                         auth_resp.content)

    def test_authtoken(self):
        """
        An auth token is only good for so long, and the server can expire it
        ahead of the time the expire field suggests. So test it and if it fails
        with a 403, refresh it with another login
        """
        resp = self.httpsession.post(
            ''.join([self.ansible_api_url, '/auth']),
            headers=self.header,
            json={"access_token": self.api_token})
        if resp.status_code == 403:
            logger.debug("Auth token invalid/expired")

    def fetch_authtoken(self):
        """
        The auth token can be provided via an environment variable or come
        through the ~/.ndnauth file. Test the token and return it if it passes
        the auth check
        """
        if self.api_token:
            logger.debug('API token already set, returning existing value')
        elif not exists(self.token_file):
            logger.warning('Auth token file not found at %s' % self.token_file)
            logger.debug('Attempting authentication via environment variables')
            self.api_auth()
            # if this file still doesn't exist, then we failed and have no choice but to give up
            if not exists(self.token_file):
                logger.warning(
                    'Auth token file not found after attempting login')
        else:
            logger.debug('Loading API token from auth file %s' %
                         self.token_file)
            self.api_token = open(self.token_file, 'r').read().strip()
            self.header['Authorization'] = 'Bearer %s' % self.api_token
            logger.debug('Loaded API token: %s' % self.api_token)

        return self.api_token

    def _get(self, path='/', params={}, headers={}, verify=True, retry_attempt=0):
        logger.debug('GET-ing URL %s' %
                     ''.join([self.ansible_api_url, path]))
        logger.debug('  with params %s' % params)
        get_headers = self.header
        get_headers.update(headers)
        logger.debug('  with headers %s' % get_headers)
        try:
            response = self.httpsession.get(''.join([self.ansible_api_url, path]),
                                            params=params,
                                            headers=get_headers,
                                            verify=verify
                                            )
            logger.debug('Received API response code: %s' %
                         response.status_code)
            response.raise_for_status()
        except Exception as err:
            # update token and try again
            if response.status_code == 403 and retry_attempt < 1:
                logger.warning(
                    'Received 403, attempting to refresh token and retrying...')
                self.api_auth()
                # Retry after updating the token
                return self._get(path, params, headers, verify, retry_attempt=1)
            else:
                logger.warn('Failed to GET data from API: %s' % err)
                raise Exception("ansible API error %s" % err)

        try:
            logger.debug('Returning API json: %s' % response.json())
            return response.json()
        except Exception as e:
            raise Exception("not able to parse json %s" % e)

    def _put(self, path='/', data=None, params={}, headers={}, verify=True, retry_attempt=0):
        logger.debug('PUT-ing URL %s' %
                     ''.join([self.ansible_api_url, path]))
        logger.debug('  with params %s' % params)
        data = json.loads(json.dumps(data, default=str))
        logger.debug('  with json %s' % data)
        put_headers = self.header
        put_headers.update(headers)
        logger.debug('  with headers %s' % put_headers)
        try:
            response = self.httpsession.put(''.join([self.ansible_api_url, path]),
                                            params=params,
                                            headers=put_headers,
                                            json=data,
                                            verify=verify
                                            )
            logger.debug('Received API response code: %s' %
                         response.status_code)
            response.raise_for_status()
        except Exception as err:
            # update token and try again
            if response.status_code == 403 and retry_attempt < 1:
                logger.warning(
                    'Received 403, attempting to refresh token and retrying...')
                self.api_auth()
                # Retry after updating the token
                return self._put(path, data, params, headers, verify, retry_attempt=1)
            else:
                logger.warn('Failed to PUT data to API: %s' % err)
                raise Exception("ansible API error %s" % err)

        try:
            logger.debug('Returning API json: %s' % response.json())
            return response.json()
        except Exception as e:
            raise Exception("not able to parse json %s" % e)

    def _post(self, path='/', data=None, params={}, headers={}, verify=True, retry_attempt=0):
        logger.debug('POST-ing URL %s' %
                     ''.join([self.ansible_api_url, path]))
        logger.debug('  with params %s' % params)
        data = json.loads(json.dumps(data, default=str))
        logger.debug('  with json %s' % data)
        post_headers = self.header
        post_headers.update(headers)
        logger.debug('  with headers %s' % post_headers)
        try:
            response = self.httpsession.post(''.join([self.ansible_api_url, path]),
                                             params=params,
                                             headers=post_headers,
                                             json=data,
                                             verify=verify
                                             )
            logger.debug('Received API response code: %s' %
                         response.status_code)
            response.raise_for_status()
        except Exception as err:
            # update token and try again
            if response.status_code == 403 and retry_attempt < 1:
                logger.warning(
                    'Received 403, attempting to refresh token and retrying...')
                self.api_auth()
                # Retry after updating the token
                return self._post(path, data, params, headers, verify, retry_attempt=1)
            else:
                logger.warn('Failed to POST data to API: %s' % err)
                raise Exception("ansible API error %s" % err)

        try:
            logger.debug('Returning API json: %s' % response.json())
            return response.json()
        except Exception as e:
            raise Exception("not able to parse json %s" % e)

    def _delete(self, path='/', params={}, headers={}, verify=True, retry_attempt=0):
        logger.debug('DELETE-ing URL %s' %
                     ''.join([self.ansible_api_url, path]))
        logger.debug('  with params %s' % params)
        delete_headers = self.header
        delete_headers.update(headers)
        logger.debug('  with headers %s' % delete_headers)
        try:
            response = self.httpsession.delete(''.join([self.ansible_api_url, path]),
                                               params=params,
                                               headers=delete_headers,
                                               verify=verify
                                               )
            logger.debug('Received API response code: %s' %
                         response.status_code)
            response.raise_for_status()
        except Exception as err:
            # update token and try again
            if response.status_code == 403 and retry_attempt < 1:
                logger.warning(
                    'Received 403, attempting to refresh token and retrying...')
                self.api_auth()
                # Retry after updating the token
                return self._post(path, params, headers, verify, retry_attempt=1)
            else:
                logger.warn('Failed to DELETE data from API: %s' % err)
                raise Exception("ansible API error %s" % err)

        try:
            logger.debug('Returning API json: %s' % response.json())
            return response.json()
        except Exception as e:
            raise Exception("not able to parse json %s" % e)

    def inventory_refresh(self):
        """
        Trigger a refresh of the ansible inventory
        """

        self.refresh_uuid_token = self._get(
            path='/inventory/refresh')['task_id']
        return self.refresh_uuid_token

    def inventory_refresh_status(self, task_uuid=None):
        """
        Get status information about the inventory refresh job
        """

        task_uuid = task_uuid if task_uuid else self.refresh_uuid_token
        return self._get(path='/fetch_task_status/%s' % task_uuid)

    def get_inventory(self):
        """
        Fetch the ansible inventory from the API
        """

        self.inventory = self._get(path='/inventory')
        return self.inventory

    def inventory_freshness(self):
        """
        Determine how fresh the ansible inventory is
        """

        return self._get(path='/freshness')

    def run(self, details):
        """
        Send an ansible run job to the ansible API
        """
        # apply run details model
        runModel = RunModel(**details)
        if runModel.extravars and 'service_queue_id' in runModel.extravars:
            self.run_id = runModel.extravars['service_queue_id']

        result = self._post(
            path='/run', data=runModel.model_dump(exclude_unset=True))
        self.run_uuid_token = result['task_id']
        return result

    def run_status(self, task_uuid=None):
        """
        Get status information about an ansible run job
        """

        task_uuid = task_uuid if task_uuid else self.run_uuid_token
        return self._get(path='/run/status/%s' % task_uuid)

    async def follow_run_logs(self, run_id=None):
        """
        Get the log file for a run. If "follow=True", use websocket to follow along
        """
        run_id = run_id or self.run_id

        up = urlparse(PROD_API_URL)
        # the python websockets client requires the URI start with "ws" or "wss"
        wsproto = 'wss' if 'https' in up.scheme else 'ws'
        uri = f'{wsproto}://{up.netloc}{up.path}/run/tail-logs/{run_id}'

        async with websockets.connect(uri) as websocket:
            try:
                while True:
                    messages = await websocket.recv()
                    for line in messages.split('\n'):
                        print(line)
            except asyncio.CancelledError:
                pass
            except websockets.ConnectionClosed:
                pass

    def run_logs(self, run_id=None):
        """
        Fetch the run logs from a completed run job
        """
        run_id = run_id or self.run_id
        return self._get(path=f'/run/logs/{run_id}')

    def runner_playbooks(self):
        """
        Fetch a list of playbooks on the active runners
        """
        return self._get(path='/runner/playbooks')

    def runner_jobs(self, pending=False, started=False, success=False, failure=False, received=False, revoked=False):
        """
        Fetch a list of all jobs that are in the runner's message bus
        """
        filters = {
            'pending': pending,
            'started': started,
            'success': success,
            'failure': failure,
            'received': received,
            'revoked': revoked}
        return self._get(path='/runner/jobs', params=filters)

    def inventory_host_refresh(self, hosts):
        """
        Trigger a refresh of the ansible inventory for host(s) (comma separated string or list)
        """
        if isinstance(hosts, list):
            hosts = ','.join(hosts)

        self.host_refresh_uuid_token = self._get(
            path=f'/inventory/hosts_refresh/{hosts}')['task_id']
        return self.host_refresh_uuid_token

    def inventory_group_refresh(self, groups):
        """
        Trigger a refresh of the ansible inventory for group(s) (comma separated string or list)
        """
        if isinstance(groups, list):
            groups = ','.join(groups)

        self.group_refresh_uuid_token = self._get(
            path=f'/inventory/groups_refresh/{groups}')['task_id']
        return self.group_refresh_uuid_token

    def inventory_object_refresh_status(self, task_uuid=None):
        """
        Get status information about the inventory refresh job
        """

        return self._get(path=f'/inventory/refresh_status/{task_uuid}')
