import json
import os
from requests import Session
from requests.adapters import HTTPAdapter
from urllib3.util import Retry
import logging
from os.path import exists, expanduser
from getpass import getuser
from ndnsecure.auth import NdnSecureAuth

log_level = os.getenv('CLUSTER_CLIENT_LOG_LEVEL', 'WARN')
logging.basicConfig(level=getattr(logging, log_level))
logger = logging.getLogger('python-cluster-client')
logger.setLevel(level=getattr(logging, log_level))

PROD_API_URL = 'https://cluster-service-master.dh-prod-eks.kubecluster.dreamhost.io'


class ClusterBase():
    def __init__(self, url=PROD_API_URL):

        self.token_file = os.path.expanduser('~%s/.ndnauth' % getuser())
        self.cluster_api_url = os.getenv(
            'CLUSTER_API_URL', url) or PROD_API_URL
        self.header = {
            "Content-Type": "application/json",
            'User-agent': 'python-cluster-client 1.0'}

        # setup auto-retry for 5XX status codes
        retry_strat = Retry(total=3, status_forcelist=list(
            range(500, 600)), backoff_factor=2, allowed_methods=frozenset(['GET', 'POST', 'PUT', 'DELETE']))
        adapter = HTTPAdapter(max_retries=retry_strat)

        self.httpsession = Session()
        self.httpsession.mount('http://', adapter)
        self.httpsession.mount('https://', adapter)

        # try to get API token from env var otherwise go through the login process
        self.api_token = os.getenv('CLUSTER_API_TOKEN', None)
        self.api_token = self.api_token or self.fetch_authtoken()

        if self.api_token:
            self.header['Authorization'] = 'Bearer %s' % self.api_token

    def api_auth(self):
        """
        Authenticate against the cluster-service, fetching a token and storing it in ~/.ndnauth
        """
        logger.debug('Authenticating against cluster-service login endpoint')
        header = self.header
        # remove auth header, as we're trying to authenticate now
        header.pop('Authorization', None)

        # try to get user/pass from ndnsecure/vault
        try:
            logger.debug("Attempting to load user/pass from NdnSecureAuth: %s" %
                         os.environ.get('CLUSTER_API_SVC_ACCOUNT'))
            username, password = NdnSecureAuth(
                os.environ.get('CLUSTER_API_SVC_ACCOUNT')).get()
        except TypeError:
            logger.debug(
                "Failed to load user/pass from NdnSecureAuth. Using environment variables")
            # if we got nothing, check env vars
            username = os.environ.get('CLUSTER_API_USERNAME', None)
            password = os.environ.get('CLUSTER_API_PASSWORD', None)

        # if we STILL got nothing, we cannot auth
        if not username or not password:
            logger.error(
                "Unable to load username/password for API authentication")
            raise Exception(
                "Unable to load username/password for API authentication")

        auth_resp = self.httpsession.post(
            ''.join([self.cluster_api_url, '/login']),
            headers=header,
            json={"username": username, "password": password})
        logger.debug('Received API response code: %s' % auth_resp.status_code)

        auth_resp.raise_for_status()
        if auth_resp:
            logger.debug("API authentication successful!")
            self.api_token = auth_resp.json()['access_token']
            self.header['Authorization'] = 'Bearer %s' % auth_resp.json()[
                'access_token']
            logger.debug("Saving token to %s" % self.token_file)
            with open(self.token_file, 'w+') as f:
                f.write(self.api_token)
            return auth_resp
        else:
            logger.debug("API authentication API failure: %s" %
                         auth_resp.content)

    def test_authtoken(self):
        """
        An auth token is only good for so long, and the server can expire it
        ahead of the time the expire field suggests. So test it and if it fails
        with a 403, refresh it with another login
        """
        resp = self.httpsession.post(
            ''.join([self.cluster_api_url, '/auth']),
            headers=self.header,
            json={"access_token": self.api_token})
        if resp.status_code == 403:
            logger.debug("Auth token invalid/expired")

    def fetch_authtoken(self):
        """
        The auth token can be provided via an environment variable or come
        through the ~/.ndnauth file. Test the token and return it if it passes
        the auth check
        """
        if self.api_token:
            logger.debug('API token already set, returning existing value')
        elif not exists(self.token_file):
            logger.warning('Auth token file not found at %s' % self.token_file)
            logger.debug('Attempting authentication via environment variables')
            self.api_auth()
            # if this file still doesn't exist, then we failed and have no choice but to give up
            if not exists(self.token_file):
                logger.warning(
                    'Auth token file not found after attempting login')
        else:
            logger.debug('Loading API token from auth file %s' %
                         self.token_file)
            self.api_token = open(self.token_file, 'r').read().strip()
            self.header['Authorization'] = 'Bearer %s' % self.api_token
            logger.debug('Loaded API token: %s' % self.api_token)

        return self.api_token

    def _get(self, path='/', params={}, headers={}, verify=True, retry_attempt=0):
        logger.debug('GET-ing URL %s' %
                     ''.join([self.cluster_api_url, path]))
        logger.debug('  with params %s' % params)
        get_headers = self.header
        get_headers.update(headers)
        logger.debug('  with headers %s' % get_headers)
        try:
            response = self.httpsession.get(''.join([self.cluster_api_url, path]),
                                            params=params,
                                            headers=get_headers,
                                            verify=verify
                                            )
            logger.debug('Received API response code: %s' %
                         response.status_code)
            response.raise_for_status()
        except Exception as err:
            # update token and try again
            if response.status_code == 403 and retry_attempt < 1:
                logger.warning(
                    'Received 403, attempting to refresh token and retrying...')
                self.api_auth()
                # Retry after updating the token
                return self._get(path, params, headers, verify, retry_attempt=1)
            else:
                logger.warn('Failed to GET data from API: %s' % err)
                raise Exception("cluster-service API error %s" % err)

        try:
            logger.debug('Returning API json: %s' % response.json())
            return response.json()
        except Exception as e:
            raise Exception("not able to parse json %s" % e)

    def _put(self, path='/', data=None, params={}, headers={}, verify=True, retry_attempt=0):
        logger.debug('PUT-ing URL %s' %
                     ''.join([self.cluster_api_url, path]))
        logger.debug('  with params %s' % params)
        data = json.loads(json.dumps(data, default=str))
        logger.debug('  with json %s' % data)
        put_headers = self.header
        put_headers.update(headers)
        logger.debug('  with headers %s' % put_headers)
        try:
            response = self.httpsession.put(''.join([self.cluster_api_url, path]),
                                            params=params,
                                            headers=put_headers,
                                            json=data,
                                            verify=verify
                                            )
            logger.debug('Received API response code: %s' %
                         response.status_code)
            response.raise_for_status()
        except Exception as err:
            # update token and try again
            if response.status_code == 403 and retry_attempt < 1:
                logger.warning(
                    'Received 403, attempting to refresh token and retrying...')
                self.api_auth()
                # Retry after updating the token
                return self._put(path, data, params, headers, verify, retry_attempt=1)
            else:
                logger.warn('Failed to PUT data to API: %s' % err)
                raise Exception("cluster-service API error %s" % err)

        try:
            logger.debug('Returning API json: %s' % response.json())
            return response.json()
        except Exception as e:
            raise Exception("not able to parse json %s" % e)

    def _post(self, path='/', data=None, params={}, headers={}, verify=True, retry_attempt=0):
        logger.debug('POST-ing URL %s' %
                     ''.join([self.cluster_api_url, path]))
        logger.debug('  with params %s' % params)
        data = json.loads(json.dumps(data, default=str))
        logger.debug('  with json %s' % data)
        post_headers = self.header
        post_headers.update(headers)
        logger.debug('  with headers %s' % post_headers)
        try:
            response = self.httpsession.post(''.join([self.cluster_api_url, path]),
                                             params=params,
                                             headers=post_headers,
                                             json=data,
                                             verify=verify
                                             )
            logger.debug('Received API response code: %s' %
                         response.status_code)
            response.raise_for_status()
        except Exception as err:
            # update token and try again
            if response.status_code == 403 and retry_attempt < 1:
                logger.warning(
                    'Received 403, attempting to refresh token and retrying...')
                self.api_auth()
                # Retry after updating the token
                return self._post(path, data, params, headers, verify, retry_attempt=1)
            else:
                logger.warn('Failed to POST data to API: %s' % err)
                raise Exception("cluster-service API error %s" % err)

        try:
            logger.debug('Returning API json: %s' % response.json())
            return response.json(), response.status_code
        except Exception as e:
            raise Exception("not able to parse json %s" % e)

    def _delete(self, path='/', params={}, headers={}, verify=True, retry_attempt=0):
        logger.debug('DELETE-ing URL %s' %
                     ''.join([self.cluster_api_url, path]))
        logger.debug('  with params %s' % params)
        delete_headers = self.header
        delete_headers.update(headers)
        logger.debug('  with headers %s' % delete_headers)
        try:
            response = self.httpsession.delete(''.join([self.cluster_api_url, path]),
                                               params=params,
                                               headers=delete_headers,
                                               verify=verify
                                               )
            logger.debug('Received API response code: %s' %
                         response.status_code)
            response.raise_for_status()
        except Exception as err:
            # update token and try again
            if response.status_code == 403 and retry_attempt < 1:
                logger.warning(
                    'Received 403, attempting to refresh token and retrying...')
                self.api_auth()
                # Retry after updating the token
                return self._post(path, params, headers, verify, retry_attempt=1)
            else:
                logger.warn('Failed to DELETE data from API: %s' % err)
                raise Exception("cluster-service API error %s" % err)

        try:
            logger.debug('Returning API json: %s' % response.json())
            return response.json()
        except Exception as e:
            raise Exception("not able to parse json %s" % e)
