from .core import MachineServiceBase, PROD_API_URL
from .schema import MachineModel, MachineLogModel, MachineModelWithLog, inDedicatedConfugrationStatuses
from .zabbix import Zabbix
from .dreampress import Dreampress
from functools import lru_cache
from cluster_client.cluster import Cluster
from cluster_client.core import ClusterBase


PERSONALITIES = (
    'isAnsibleManaged',
    'isBackupSerf',
    'isCentralDb',
    'isDPGcpLxc',
    'isDedicated',
    'isGcpLxcHost',
    'isIptablesSkipAnsible',
    'isLocaldataBackup',
    'isLxcHost',
    'isLxcInstance',
    'isLxcMysql',
    'isLxcWeb',
    'isMail',
    'isMailFiler',
    'isNetboot',
    'isNetworkConfigAnsible',
    'isOpenstack',
    'isPeon',
    'isPrivateServer',
    'isPrivateWebServer',
    'isSSD',
    'isSerf',
    'isSharedMysql',
    'isSharedWeb',
    'isSled',
    'isVServerGuest',
    'isVSqlGuest',
    'isVSqlHost',
    'isVserverHost',
    'isWpaasMysql',
    'isWpaasShared',
    'isWpaasWeb')


class Machine():
    def __init__(self, machine={}, client=None, log_level='WARN', url=PROD_API_URL):
        self.machine = MachineModel(**machine)
        self.params = []
        self.client = client or MachineServiceBase(
            log_level=log_level, url=url)
        self.zabbix = Zabbix(self.machine.machine, client)
        self.dreampress = Dreampress(self.machine.machine, client)
        self.cluster_client = ClusterBase()
        self.inherited_clusters = [
            Cluster({'cluster': None}, self.cluster_client)]
        self.assigned_clusters = [
            Cluster({'cluster': None}, self.cluster_client)]
        self.clusters = [Cluster({'cluster': None}, self.cluster_client)]
        self.logs = []
        self.physical = None
        self._set_personalities()

    def _set_personalities(self, personalities={}):
        for personality in PERSONALITIES:
            setattr(self, personality, personalities.get(personality, None))

    # @lru_cache()
    def get(self, machine=None, log=False, params=False, account=False, domainservices=False, extended=True):
        """
        Fetch the machine details from the API
          log: include the machine action logs
          params: include machine parameters
          account:
          domainservices:
          extended: include all the personalities (i.e. isSled) as well as clusters
        """
        machine = machine or self.machine.machine
        get_params = {'log': log, 'params': params, 'account': account,
                      'domainservices': domainservices, 'extended': extended}
        result = self.client._get(path='/machine/%s' %
                                  machine, params=get_params)

        if log:
            self.machine = MachineModel(**result)
            self.log = [MachineLogModel(**log)
                        for log in result.get('log', [])]
        else:
            self.machine = MachineModel(**result)

        if extended:
            self._set_personalities(result)
            self.inherited_clusters = [Cluster(
                {'cluster': cluster}, self.cluster_client) for cluster in result.get('inherited_clusters', [])]
            self.assigned_clusters = [Cluster(
                {'cluster': cluster}, self.cluster_client) for cluster in result.get('assigned_clusters', [])]
            self.clusters = [Cluster({'cluster': cluster}, self.cluster_client)
                             for cluster in result.get('clusters', [])]
            self.physical = result.get(
                'physical', None) if result['realm'] == 'logical' else None

        if params:
            self.params = result.get('params', [])

        return self.machine

    # @lru_cache()
    def vps_version(self, machine=None):
        machine = machine or self.machine.machine
        return self.client._get(path='/machine/%s/VPSVersion' % machine)

    # @lru_cache()
    def vps_preorder(self, machine=None):
        machine = machine or self.machine.machine
        return self.client._get(path='/machine/%s/VPSPreorder' % machine)

    # @lru_cache()
    def is_wpaas_web(self, machine=None):
        machine = machine or self.machine.machine
        return self.client._get(path='/machine/%s/isWpaasWeb' % machine)

    # @lru_cache()
    def is_private_webserver(self, machine=None):
        machine = machine or self.machine.machine
        return self.client._get(path='/machine/%s/isPrivateWebServer' % machine)

    # @lru_cache()
    def is_vserver_guest(self, machine=None):
        machine = machine or self.machine.machine
        return self.client._get(path='/machine/%s/isVserverGuest' % self.machine)

    # @lru_cache()
    def is_openstack(self, machine=None):
        machine = machine or self.machine.machine
        return self.client._get(path='/machine/%s/isOpenstack' % machine)

    # @lru_cache()
    def is_lxc_instance(self, machine=None):
        machine = machine or self.machine.machine
        return self.client._get(path='/machine/%s/isLxcInstance' % machine)

    # @lru_cache()
    def is_lxc_host(self, machine=None):
        machine = machine or self.machine.machine
        return self.client._get(path='/machine/%s/isLxcHost' % machine)

    # @lru_cache()
    def lxc_instances(self, machine=None):
        machine = machine or self.machine.machine
        return self.client._get(path='/machine/%s/lxc/instances' % machine)

    # @lru_cache()
    def domains(self, machine=None):
        machine = machine or self.machine.machine
        return self.client._get(path='/machine/%s/domains' % machine)

    # @lru_cache()
    def search(self, machine=None):
        machine = machine or self.machine.machine
        return self.client._get(path='/machine/search/%s' % machine)

    # @lru_cache()
    def search_or_get(self, machine=None):
        machine = machine or self.machine.machine
        return self.client._get(path='/machine/search_or_get/%s' % machine)

    # @lru_cache()
    def param(self, machine=None):
        machine = machine or self.machine.machine
        return self.client._get(path='/machine/param/%s' % machine)

    # @lru_cache()
    def openstack(self, machine=None):
        machine = machine or self.machine.machine
        return self.client._get(path='/machine/openstack/%s' % machine)

    # @lru_cache()
    def services(self, machine=None):
        machine = machine or self.machine.machine
        return self.client._get(path='/machine/%s/services' % machine)

    # @lru_cache()
    def kentik(self, machine=None):
        machine = machine or self.machine.machine
        return self.client._get(path='/kentik/machine/%s' % machine)

    # @lru_cache()
    def ip(self, machine=None):
        machine = machine or self.machine.machine
        return self.client._get(path='/ip/machine/%s' % machine)

    # @lru_cache()
    def save(self):
        return self.client._post(path='/machine', data=self.machine.model_dump())

    # @lru_cache()
    def down(self):
        self.machine.status = 'down'
        self.save()
        return self.machine

    # @lru_cache()
    def get_params(self, machine=None):
        machine = machine or self.machine.machine
        self.get(machine, params=True)
        return self.params

    def get_param(self, parameter):
        return [param for param in self.params if param['parameter'] == parameter]

    def set_param(self, parameter, value, machine=None):
        machine = machine or self.machine.machine
        self.client._post(
            path=f'/machine/param/{machine}', params={'key': parameter, 'value': value})
        self.get_params()
        return self.get_param(parameter)

    def search(self, log=False):
        return [MachineModel(**result) for result in self.client._post(path=f'/machine/search', data=self.machine.model_dump(exclude_none=True), params={'log': log})]

    def dedicated_configuration_by_status(self, status):
        data = None
        if isinstance(status, str):
            inDedicatedConfugrationStatuses(**{'status': status})
            data = {'status': [status]}
        elif isinstance(status, list):
            [inDedicatedConfugrationStatuses(
                **{'status': stat}) for stat in status]
            data = {'status': status}
        elif isinstance(status, dict):
            # dict must have "status" key
            if status.get('status', False):
                if isinstance(status['status'], str):
                    inDedicatedConfugrationStatuses(**status)
                    data = {'status': [status.get('status')]}
                elif isinstance(status['status'], list):
                    [inDedicatedConfugrationStatuses(
                        **{'status': stat}) for stat in status['status']]
                    data = {'status': status.get('status')}
                else:
                    data = None

        return self.client._post(
            path='/machine/dedicated_configuration/by-status',
            data=data)

    def user_services(self, machine=None, active_only=False):
        machine = machine or self.machine.machine
        return self.client._get(path=f'/machine/{machine}/user_services', params={'active_only': active_only})
