from cement import Controller, ex
import hvac
from functools import wraps
from datetime import datetime
import getpass
import re
import os


#  Decorator to ensure that the user is authenticated before running a function
def ensure_authenticated(func):
    @wraps(func)
    def wrapper(controller_instance, *args, **kwargs):
        vault_client = controller_instance.app.get_vault_client()
        if not vault_client.is_authenticated():
            controller_instance.app.log.error(
                "Authentication failed. Please login with `ndn vault login`.")
            controller_instance.app.exit_code = 1
            controller_instance.app.close()
            return
        return func(controller_instance, vault_client, *args, **kwargs)
    return wrapper


class AlertSilence(Controller):
    class Meta:
        label = 'alert_silence'
        stacked_type = 'nested'
        stacked_on = 'base'
        help = "Manage alerts silences"
        extensions = ['tabulate']
        output_handler = 'tabulate'
        arguments = [
            (['-T', '--token'],
                {'help': 'The name of the vault secret from which to load the API token',
                 'default': 'py-microservice-default-token',
                 'action': 'store'})
        ]

    #  Constructor to warn user if running as root
    def __init__(self, *args, **kwargs):
        super(AlertSilence, self).__init__(*args, **kwargs)
        if os.geteuid() == 0:
            print(
                "This script should not be run as root. Please run as a non-privileged user.")

    def _fetch_token_from_vault(self, vault_client, token=None):
        token = token or self.app.pargs.token
        try:
            content = vault_client.secrets.kv.v1.read_secret(
                path='auth/%s' % token, mount_point='secret')['data']['content'].split()[1]
        except hvac.exceptions.InvalidPath:
            self.app.log.warning(
                "Token content not found in vault for secret name %s" % token)
            content = None
        return content

    def _api_call(self, vault_client, method, endpoint='', data=None, params=None):
        headers = {
            'Authorization': f'Bearer {self._fetch_token_from_vault(vault_client)}'}
        base_url = self.app.config.get('ndn', 'alert_silence_api_host')
        url = f"{base_url}{endpoint}"

        try:
            if method == 'get':
                response_data = self.app.get_url(
                    url, headers=headers, params=params)
            elif method == 'post':
                response_data = self.app.post_json_url(
                    url, headers=headers, params=params)
            elif method == 'delete':
                response_data = self.app.delete_json_url(
                    url, headers=headers, params=params)

            return response_data

        except Exception as e:
            self.app.log.error(
                f'API call failed due to an exception: {str(e)}')
            return None

    #  for some reason at some point I was seeing dates in different formats. Ideally not necessary.
    def _format_date(self, date_str):
        """Helper function to format date strings with multiple possible formats."""
        date_formats = ['%Y-%m-%dT%H:%M:%S.%fZ', '%Y-%m-%dT%H:%M:%SZ']

        for fmt in date_formats:
            try:
                return datetime.strptime(date_str, fmt).strftime('%Y-%m-%d %H:%M:%S')
            except ValueError:
                continue
        raise ValueError(
            f"time data {date_str} does not match expected formats")

    @ex(
        help='Show alert silences based on ID, machine, or show all.'
             'Note: Certain flags are only applicable in specific contexts.',
        arguments=[
            (['identifier'],
             {'help': 'UUID of the alert silence or machine name, or omit to list all',
              'action': 'store',
              'nargs': '?',
              'default': None}),
            (['--include-expired'],
             {'help': 'Include expired silences in the list (applicable when showing all or by machine)',
              'default': True,
              'action': 'store_false'}),
            (['--include-logical'],
             {'help': 'Include silences for logical machine if a physical vmhost is passed (applicable only with machine name)',
              'action': 'store_true'}),
            (['--include-host'],
             {'help': 'Include silences for vmhost if a VM is passed (applicable only with machine name)',
              'action': 'store_true'}),
            (['-m', '--match'],
             {'help': 'List of Prometheus matchers in the format {key}{operator}{value}, e.g., instance=~^test (applicable when showing all)',
              'default': None,
              'action': 'store',
              'nargs': '+'})
        ]
    )
    @ensure_authenticated
    def show(self, vault_client):
        identifier = self.app.pargs.identifier
        params = {
            "humanreadable": True,
            "activeonly": self.app.pargs.include_expired,
            "includelogical": self.app.pargs.include_logical,
            "includehost": self.app.pargs.include_host,
            "matchers": self.app.pargs.match
        }

        #  Detect if the identifier is a UUID or a machine name
        if identifier:
            if re.match(r'^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$', identifier, re.I):
                endpoint = f"/alert/silences/id/{identifier}"
            else:
                endpoint = f"/alert/silences/machine/{identifier}"
        else:
            endpoint = "/alert/silences"

        response = self._api_call(vault_client, 'get', endpoint, params=params)

        if response:
            if isinstance(response, dict):
                response = [response]

            for item in response:
                item['status'] = item['status'].get('state', 'unknown')
                item['matchers'] = "\n".join(item['matchers'].split())
                item['startsAt'] = self._format_date(item['startsAt'])
                item['endsAt'] = self._format_date(item['endsAt'])
                item['updatedAt'] = self._format_date(item['updatedAt'])

            self.app.render(response, headers='keys', tablefmt='rounded_grid')
        else:
            self.app.log.error(
                "No Silences found or error occurred. Use -d to possibly see more details.")
            self.app.exit_code = 1
            self.app.close()

        return response

    @ex(
        help='Create an alert silence',
        arguments=[
            (['-m', '--machine'],
             {'help': 'The machine to apply the silence to', 'required': True}),
            (['-D', '--duration'],
             {'help': 'Duration of the silence, e.g., "1h" or "7d"', 'required': True}),
            (['-c', '--comment'],
             {'help': 'Comment about why the silence is being created', 'required': True}),
            (['-C', '--created-by'],
             {'help': 'Username of the person creating the silence (if not set, will evaluate current user ID)', 'required': False}),
            (['-r', '--recursive'],
             {'help': 'Include silences for the machine\'s children', 'action': 'store_true', 'default': False}),
            (['-a', '--alertname'],
             {'help': 'Specific alert name to silence', 'required': False}),
        ]
    )
    @ensure_authenticated
    def create(self, vault_client):
        alertname = self.app.pargs.alertname
        endpoint = f"/alert/silences/{self.app.pargs.machine}/alertname/{alertname}" \
            if alertname else f"/alert/silences/{self.app.pargs.machine}"

        created_by = self.app.pargs.created_by or (
            getpass.getuser() if os.geteuid() != 0
            else input("Since you are running as root, please enter the username of the person creating this silence: ")
        )

        params = {
            "duration": self.app.pargs.duration,
            "createdBy": created_by,
            "comment": self.app.pargs.comment,
            "recursive": self.app.pargs.recursive
        }

        response = self._api_call(
            vault_client, 'post', endpoint, params=params)

        if response:
            silence_ids = [item['silenceID']
                           for item in response if 'silenceID' in item]

            if silence_ids:
                silence_ids_str = ', '.join(silence_ids)
                print(
                    f'Silence(s) created successfully with IDs: {silence_ids_str}')
            else:
                self.app.log.error('No silence IDs found in the response.')
                self.app.exit_code = 1
                self.app.close()

        else:
            self.app.log.error(
                'Failed to create silence. Use -d to possibly see more details.')
            self.app.exit_code = 1
            self.app.close()

            return response

    @ex(
        help='Expire alert silences based on one or more UUIDs',
        arguments=[
            (['uuids'],
             {'help': 'Space-separated UUIDs of the alert silences to expire',
              'action': 'store',
              'nargs': '+',
              'type': str})
        ]
    )
    @ensure_authenticated
    def expire(self, vault_client):
        silence_ids = self.app.pargs.uuids

        expired_ids = []
        not_found_ids = []
        inactive_ids = []

        for silence_id in silence_ids:
            silence_id = silence_id.strip()  # Clean up any extra spaces
            existing = self._api_call(
                vault_client, 'get', f"/alert/silences/id/{silence_id}")

            if not existing:
                not_found_ids.append(silence_id)
                continue

            if existing['status']['state'] != 'active':
                inactive_ids.append(silence_id)
                continue

            response = self._api_call(
                vault_client, 'delete', f"/alert/silences/id/{silence_id}")

            if response:
                expired_ids.append(silence_id)
            else:
                self.app.log.error(
                    f"Failed to expire silence ID {silence_id}. Use -d to possibly see more details.")
                self.app.exit_code = 1
                self.app.close()
                return

        if expired_ids:
            print(f'Successfully expired silences: {", ".join(expired_ids)}')
        if not_found_ids:
            print(f'Silences not found: {", ".join(not_found_ids)}')
        if inactive_ids:
            print(
                f'Silences not active - no action taken: {", ".join(inactive_ids)}')

        if not expired_ids and not not_found_ids and not inactive_ids:
            print(
                "No actions were taken. Check the UUIDs provided or run with -d for details.")
