from cement import Controller, ex
from pprint import pprint
from ansible_client.base import AnsibleClient
import asyncio
import time
from getpass import getuser


class Ansible(Controller):
    class Meta:
        label = 'ansible'
        stacked_type = 'nested'
        stacked_on = 'base'
        help = 'Ansible-based administrative tools'

    def _follow_run_log(self, aclient, run_id):
        self.app.log.debug("Following logs using a websocket")

        async def do(aclient, run_id):
            await aclient.follow_run_logs(run_id)
        loop = asyncio.get_event_loop()
        try:
            self.app.log.debug("Looping until client hits Ctlr+c")
            loop.run_until_complete(do(aclient, run_id))
        except KeyboardInterrupt:
            self.app.log.debug("Keyboard interrupt dectected")
            return

    @ex(help='Schedule an ansible run in the ansible-jobs runner. This will return a "task_id" which is the celery task UUID and a "run_id" which is '
             'the ansible job UUID. You can probably ignore the "task_id".',
        arguments=[
            (['playbook'],
             {'help': 'Playbook to run (relative path to top of ansible repo)',
              'action': 'store'}),
            (['-t', '--tags'],
             {'help': 'Playbook tags to explicily include',
              'action': 'store'}),
            (['-s', '--skip_tags'],
             {'help': 'Playbook tags to explicitly skip',
              'action': 'store'}),
            (['-l', '--limit'],
             {'help': 'Limit the play to specific hosts/groups',
              'action': 'store'}),
            (['-H', '--host_pattern'],
             {'help': 'Host pattern to targer during playbook execution',
              'action': 'store'}),
            (['-V', '--verbosity'],
             {'help': 'Log verbosity',
              'action': 'store'}),
            (['-e', '--extra_vars'],
             {'help': 'Extra variables to include during the playbook execution',
              'type': str,
              'default': [],
              'nargs': '*',
              'action': 'store'}),
            (['-f', '--forks'],
             {'help': 'Number of forks to use for the playbook execution',
              'action': 'store'}),
            (['-u', '--user'],
             {'help': 'Username for ansible to use for ssh-ing (default: ansible)',
              'action': 'store',
              'default': 'ansible'}),
            (['-U', '--dh_ansible_user'],
             {'help': 'Username used to report metrics as to who requested this playbook to run',
              'action': 'store',
              'default': getuser()}),
            (['-F', '--follow'],
             {'help': 'Follow the run logs after scheduling the run',
              'action': 'store_true',
              'default': False}),
        ])
    def run(self):
        self.app.require_token()
        aclient = AnsibleClient()
        # parse extravars
        xv = {}
        for i in self.app.pargs.extra_vars:
            try:
                key, value = i.split('=', 1)
                xv[key] = value
            except:  # noqa
                self.app.log.warning(
                    "extra_vars must be in the form of key=value")
                self.app.exit_code = 1
                self.app.close()
        self.app.pargs.extra_vars = xv
        result = aclient.run({
            'playbook': self.app.pargs.playbook,
            'tags': self.app.pargs.tags,
            'skip_tags': self.app.pargs.skip_tags,
            'limit': self.app.pargs.limit,
            'host_pattern': self.app.pargs.host_pattern,
            'verbosity': self.app.pargs.verbosity,
            'extravars': self.app.pargs.extra_vars,
            'forks': self.app.pargs.forks,
            'dh_ansible_user': self.app.pargs.dh_ansible_user,
            'ansible_remote_user': self.app.pargs.user,
        })
        if result:
            if self.app.pargs.follow:
                # sleep for a few seconds to let the worker pick the job up
                time.sleep(3)
                self.app.log.info(
                    "Following logs for run_id: %s" % result['run_id'])
                self._follow_run_log(aclient, result['run_id'])
            else:
                self.app.log.info("Successfully scheduled playbook run: %s" %
                                  result)
        else:
            self.app.log.warn(
                "Failed to schedule an ansible playbook run. Try again with -d for more information")

    @ex(help='Fetch the status of a celery task for ansible jobs (use the task_id, not the run_id)',
        arguments=[
            (['task_uuid'],
             {'help': 'UUID of celery task',
              'action': 'store'}),
        ])
    def status(self):
        self.app.require_token()
        aclient = AnsibleClient()
        try:
            result = aclient.run_status(self.app.pargs.task_uuid)
        except Exception:
            result = None
            self.app.log.warning(
                "Failed to get task status for UUID: %s" % self.app.pargs.task_uuid)
            self.app.log.warning(
                "Did you use the run_id instead of the task_id?")
        if result:
            pprint(result)

    @ex(help='Fetch logs of an ansible run job',
        arguments=[
            (['run_id'],
             {'help': 'ID of an ansible run (usuall service_queue_id from extravars field)',
              'action': 'store'}),
            (['-f', '--follow'],
             {'help': 'Tail the log',
              'action': 'store_true',
              'default': False}),
        ])
    def logs(self):
        self.app.require_token()
        aclient = AnsibleClient()

        if self.app.pargs.follow:
            self._follow_run_log(aclient, self.app.pargs.run_id)
        else:
            try:
                result = aclient.run_logs(self.app.pargs.run_id)
                if result:
                    print(result)
            except Exception as e:
                self.app.log.info(
                    "Failed to get logs for job ID: %s" % self.app.pargs.run_id)
                self.app.log.info(
                    "Did you use the task_id instead of the run_id?")
                self.app.log.info(
                    'If the ansible job is still running, try again with the "-f" flag')
                self.app.log.info("Re-run with '-d' flag for more information")
                self.app.log.debug(e)
                self.app.exit_code = 1
                self.app.close()

    @ex(help='Fetch list of playbooks on the runner(s)',
        arguments=[
            (['--one-line'],
             {'help': 'List all playbooks on a single line, comma separated',
              'action': 'store_true',
              'default': False}),
        ])
    def playbooks(self):
        self.app.require_token()
        aclient = AnsibleClient()

        result = aclient.runner_playbooks()
        if self.app.pargs.one_line:
            print(', '.join(sorted(result)))
        else:
            print('\n'.join(sorted(result)))

    @ex(help='Fetch list of ansible jobs registered in the celery queue',
        arguments=[
            (['--success'],
             {'help': 'Filter for jobs that were successful',
              'action': 'store_true',
              'default': False}),
            (['--failure'],
             {'help': 'Filter for jobs that failed',
              'action': 'store_true',
              'default': False}),
            (['--received'],
             {'help': 'Filter for jobs that have been received',
              'action': 'store_true',
              'default': False}),
            (['--revoked'],
             {'help': 'Filter for jobs that have been revoked',
              'action': 'store_true',
              'default': False}),
            (['--started'],
             {'help': 'Filter for jobs that have been started',
              'action': 'store_true',
              'default': False}),
        ])
    def jobs(self):
        self.app.require_token()
        aclient = AnsibleClient()

        print(aclient.runner_jobs(
            success=self.app.pargs.success,
            failure=self.app.pargs.failure,
            received=self.app.pargs.received,
            revoked=self.app.pargs.revoked,
            started=self.app.pargs.started,
        ))
