from cement import Controller, ex
import subprocess
import os
import json


class AnsibleUtils(Controller):
    class Meta:
        label = 'utils'
        stacked_type = 'nested'
        stacked_on = 'ansible'
        help = 'Various helper utilities for ansible'
        arguments = [
            (['-c', '--config'],
             {'help': 'Full path to your ansible config file (default ~/.ansible.cfg)',
             'action': 'store',
              'default': None}),
        ]

    def _run_ansible_cmd(self, cmd):
        self.app.log.debug("Running ansible command: %s" % ' '.join(cmd))
        result = subprocess.run(cmd,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                check=True,
                                text=True)
        self.app.log.debug("ansible command stderr: %s" % result.stderr)
        return result

    @ex(help='Prints all groups for a given host',
        arguments=[
            (['host'],
             {'help': 'Hostname of an ansible machine',
              'action': 'store'}),
            (['-s', '--short'],
             {'help': 'Short-form output (all on one line, space separated)',
              'action': 'store_true',
              'default': False}),
        ])
    def groups(self):
        if self.app.pargs.config:
            os.environ['ANSIBLE_CONFIG'] = self.app.pargs.config
        cmd = ['ansible', '-o', self.app.pargs.host,
               '-m', 'debug', '-a', 'var=group_names']
        agroups = self._run_ansible_cmd(cmd)
        result = agroups.stdout
        raw = result.split('=>')[1]
        groups = json.loads(raw).get('group_names', [])
        if self.app.pargs.short:
            print(' '.join(groups))
        else:
            print('\n'.join(groups))

    @ex(help='Prints all groups containing the given string',
        arguments=[
            (['search_string'],
             {'help': 'String contained within the group name',
              'action': 'store'}),
            (['-s', '--short'],
             {'help': 'Short-form output (all on one line, space separated)',
              'action': 'store_true',
              'default': False}),
        ])
    def group_search(self):
        if self.app.pargs.config:
            os.environ['ANSIBLE_CONFIG'] = self.app.pargs.config
        cmd = ['ansible', '-o', 'localhost', '-m', 'debug', '-a', 'var=groups']
        agroups = self._run_ansible_cmd(cmd)
        result = agroups.stdout
        raw = result.split('=>')[1]
        all_groups = [group for group in json.loads(
            raw).get('groups', {}).keys()]
        self.app.log.debug("found %s groups" % len(all_groups))
        groups = [
            group for group in all_groups if self.app.pargs.search_string in group]
        if self.app.pargs.short:
            print(' '.join(groups))
        else:
            print('\n'.join(groups))

    @ex(help='Prints all hosts in a gvien group (and subgroups)',
        arguments=[
            (['group'],
             {'help': 'String contained within the group name',
              'action': 'store'}),
            (['-s', '--short'],
             {'help': 'Short-form output (all on one line, space separated)',
              'action': 'store_true',
              'default': False}),
        ])
    def hosts_in_group(self):
        if self.app.pargs.config:
            os.environ['ANSIBLE_CONFIG'] = self.app.pargs.config
        cmd = ['ansible-inventory', '--graph', self.app.pargs.group]
        groups = self._run_ansible_cmd(cmd)
        result = groups.stdout
        parsed_result = result.replace('|', '').replace(
            '--', '').replace(':', '').split()
        hosts = [host for host in parsed_result if not host.startswith('@')]
        if self.app.pargs.short:
            print(' '.join(hosts))
        else:
            print('\n'.join(hosts))
