from cement import Controller, ex
import hvac
from shutil import which
from subprocess import call


class Db(Controller):
    class Meta:
        label = 'db'
        stacked_type = 'nested'
        stacked_on = 'base'
        help = 'Inspect information about DB connection information and launch an interactive DB client'

    def _fetch_db_from_vault(self, vclient, db):
        try:
            content = vclient.secrets.kv.v1.read_secret(
                path='db/%s' % db, mount_point='secret')['data']
        except hvac.exceptions.InvalidPath:
            self.app.log.warning(
                "DB information not found in vault for db %s" % db)
            content = None
        return content

    @ex(
        help='Launch an interactive DB client',
        arguments=[
            (['database'],
             {'help': 'Database name to connect to',
              'action': 'store',
              'default': None}),
            (['-e', '--execute'],
             {'help': 'SQL to execute and then exit',
              'action': 'store',
              'default': None})
        ]
    )
    def connect(self):
        vclient = self.app.get_vault_client()
        if not vclient.is_authenticated():
            self.app.exit_code = 1
            self.app.close()

        dbi = self._fetch_db_from_vault(vclient, self.app.pargs.database)
        if not dbi:
            # not in vault. maybe it's a customer DB? :shrug:
            self.app.log.error("Unable to locate db %s" %
                               self.app.pargs.database)
            self.app.exit_code = 1
            self.app.close()
        # we're just going to assume mysql for now. support for other clients can be added later
        mysql_client = which('mysql') or which('mysql-client')
        if not mysql_client:
            self.app.log.error('Unable to locate mysql client in $PATH')
            self.app.exit_code = 1
            self.app.close()

        cmd = mysql_client + " -A --host='%s' --user='%s' --password='%s' %s" % (
            dbi['host'], dbi['username'], dbi['password'], dbi['database'])
        if self.app.pargs.execute:
            cmd = cmd + ' -e "%s"' % self.app.pargs.execute
        self.app.log.debug('Executing shell command: %s' % cmd)
        call(cmd, shell=True)

    @ex(
        help='Display DB connection information',
        arguments=[
            (['database'],
             {'help': 'Database name to fetch info about',
              'action': 'store'}),
        ]
    )
    def info(self):
        vclient = self.app.get_vault_client()
        if not vclient.is_authenticated():
            self.app.exit_code = 1
            self.app.close()

        result = self._fetch_db_from_vault(vclient, self.app.pargs.database)
        if not result:
            # not in vault. maybe it's a customer DB? :shrug:
            self.app.log.error("Unable to locate db %s" %
                               self.app.pargs.database)
            self.app.exit_code = 1
            self.app.close()
        result['args'] = self.app.pargs.__dict__
        self.app.render({'data': result}, 'db/info.jinja2')

    @ex(help='Display a list of DBs with connection details in vault')
    def list(self):
        vclient = self.app.get_vault_client()
        if not vclient.is_authenticated():
            self.app.exit_code = 1
            self.app.close()

        dbs = vclient.secrets.kv.v1.list_secrets(path='db/', mount_point='secret')['data']['keys']
        self.app.render({'data': dbs}, 'db/list.jinja2')
