from cement import Controller, ex
from getpass import getpass
from datetime import datetime
import os
import pprint
from urllib.parse import urlparse


class Gitlab(Controller):
    class Meta:
        label = 'gitlab'
        stacked_type = 'nested'
        stacked_on = 'base'
        help = 'Gitlab operations'

    @ex(help='Display web UI URL for Gitlab')
    def url(self):
        urlp = urlparse(self.app.config.get('ndn', 'gitlab_api'))
        print(f"{urlp.scheme}://{urlp.hostname}/")


class GitlabPersonal(Controller):
    class Meta:
        label = 'personal'
        stacked_type = 'nested'
        stacked_on = 'gitlab'
        help = 'Gitlab personal token operations'


class GitlabPersonalToken(Controller):
    class Meta:
        label = 'token'
        stacked_type = 'nested'
        stacked_on = 'personal'
        help = 'Gitlab personal token operations'

    @ex(
        help='Create a Gitlab personal token and save it to a file (~/.gitlab-token-ndn by default)',
        arguments=[
            (['-e', '--email'],
             {'help': 'Gitlab email',
              'action': 'store',
              'default': None}),
            (['-p', '--password'],
             {'help': 'Gitlab password',
              'action': 'store',
              'default': None}),
            (['-f', '--file'],
             {'help': 'File location to store this access token',
              'action': 'store',
              'default': '~/.gitlab-token-ndn'}),
        ]
    )
    def create(self):
        if not self.app.pargs.email:
            self.app.log.error('Please provide your Gitlab email')
            self.app.exit_code = 1
            self.app.close()

        # get an access token so we can interact with the Gitlab API
        access_token = self.app.post_json_url(
            # auth is against the gitlab host, not the API
            self.app.config.get('ndn', 'gitlab_api_host') + "/oauth/token",
            json={
                'grant_type': 'password',
                'username': self.app.pargs.email,
                'password': self.app.pargs.password or getpass()})
        if not access_token:
            self.app.log.error(
                'Failed to authenticate with Gitlab. Re-run with -d flag to get more details.')
            self.app.exit_code = 1
            self.app.close()

        # fetch our user ID
        user_id = self.app.get_url(
            self.app.config.get('ndn', 'gitlab_api') + '/users',
            params={'search': self.app.pargs.email},
            headers={'Authorization': '%s %s' % (access_token['token_type'], access_token['access_token'])})
        if not user_id:
            self.app.log.error(
                "So... we auth'd with this email, but now we can't seem to find a user with it? Re-run with -d flag to get more details.")
            self.app.exit_code = 1
            self.app.close()

        personal_access_token = self.app.post_json_url(
            self.app.config.get('ndn', 'gitlab_api') +
            '/users/%s/personal_access_tokens' % user_id[0]['id'],
            expected_status=201,
            headers={'Authorization': '%s %s' % (
                access_token['token_type'], access_token['access_token'])},
            json={
                'name': 'python-ndn-cli-%s-%s-%s' % (datetime.now().year, datetime.now().month, datetime.now().day),
                'scopes': ['read_repository', 'read_registry']
            })
        if not personal_access_token:
            self.app.log.error(
                "Failed to create a personal access token. Re-run with -d flag to get more details.")
            self.app.exit_code = 1
            self.app.close()

        tfile_path = os.path.expanduser(self.app.pargs.file)
        if os.path.isfile(tfile_path):
            overwrite = input(
                "File %s already exists. Overwrite? Y/N" % tfile_path)
            if overwrite.lower() == 'y':
                self.app.log.debug(
                    "Removing file %s at users request" % tfile_path)
                os.remove(tfile_path)
            else:
                self.app.log.error(
                    "Not overriding existing token file at %s" % tfile_path)
                self.app.exit_code = 1
                self.app.close()

        with open(tfile_path, 'w+') as f:
            f.write(personal_access_token['token'])

        print(
            'Successfully fetched and wrote Gitlab personal access token to %s' % tfile_path)

    @ex(
        help='List gitlab personal access tokens',
        arguments=[
            (['-e', '--email'],
             {'help': 'Gitlab email',
              'action': 'store',
              'default': None}),
            (['-p', '--password'],
             {'help': 'Gitlab password',
              'action': 'store',
              'default': None}),
        ]
    )
    def list(self):
        if not self.app.pargs.email:
            self.app.log.error('Please provide your Gitlab email')
            self.app.exit_code = 1
            self.app.close()

        # get an access token so we can interact with the Gitlab API
        access_token = self.app.post_json_url(
            # auth is against the gitlab host, not the API
            self.app.config.get('ndn', 'gitlab_api_host') + "/oauth/token",
            json={
                'grant_type': 'password',
                'username': self.app.pargs.email,
                'password': self.app.pargs.password or getpass()})
        if not access_token:
            self.app.log.error(
                'Failed to authenticate with Gitlab. Re-run with -d flag to get more details.')
            self.app.exit_code = 1
            self.app.close()

        # fetch our user ID
        user_id = self.app.get_url(
            self.app.config.get('ndn', 'gitlab_api') + '/users',
            params={'search': self.app.pargs.email},
            headers={'Authorization': '%s %s' % (access_token['token_type'], access_token['access_token'])})
        if not user_id:
            self.app.log.error(
                "So... we auth'd with this email, but now we can't seem to find a user with it? Re-run with -d flag to get more details.")
            self.app.exit_code = 1
            self.app.close()

        tokens = self.app.get_url(
            self.app.config.get('ndn', 'gitlab_api') +
            '/personal_access_tokens',
            params={'user_id': user_id[0]['id']},
            headers={'Authorization': '%s %s' % (access_token['token_type'], access_token['access_token'])})
        if not tokens:
            self.app.log.error(
                "We didn't seem to find any tokens. Perhaps you simply do not have any? Re-run with -d flag to get more details.")
            self.app.exit_code = 1
            self.app.close()
        else:
            pprint.pprint(tokens)

    @ex(
        help='Delete a Gitlab personal access token',
        arguments=[
            (['token_id'],
             {'help': 'ID of the personal access token to delete',
              'action': 'store',
              'default': None}),
            (['-e', '--email'],
             {'help': 'Gitlab email',
              'action': 'store',
              'default': None}),
            (['-p', '--password'],
             {'help': 'Gitlab password',
              'action': 'store',
              'default': None}),
        ]
    )
    def delete(self):
        if not self.app.pargs.email:
            self.app.log.error('Please provide your Gitlab email')
            self.app.exit_code = 1
            self.app.close()

        # get an access token so we can interact with the Gitlab API
        access_token = self.app.post_json_url(
            # auth is against the gitlab host, not the API
            self.app.config.get('ndn', 'gitlab_api_host') + "/oauth/token",
            json={
                'grant_type': 'password',
                'username': self.app.pargs.email,
                'password': self.app.pargs.password or getpass()})
        if not access_token:
            self.app.log.error(
                'Failed to authenticate with Gitlab. Re-run with -d flag to get more details')
            self.app.exit_code = 1
            self.app.close()

        token_delete = self.app.httpsession.delete(
            self.app.config.get('ndn', 'gitlab_api') +
            '/personal_access_tokens/' + self.app.pargs.token_id,
            headers={'Authorization': '%s %s' % (access_token['token_type'], access_token['access_token'])})
        if token_delete.status_code != 204:
            self.app.log.error(
                "We were unable to delete this token. Re-run with -d flag to get more details.")
            self.app.exit_code = 1
            self.app.close()
        else:
            print("Successfully deleted token %s" % self.app.pargs.token_id)
