from cement import Controller, ex
import re
import ipaddress


def _isMacAddress(mac):
    """
    Is this a mac address?
    """

    if re.match("[0-9a-f]{2}([-:]?)[0-9a-f]{2}(\\1[0-9a-f]{2}){4}$", mac.lower()):
        return True
    else:
        return False


def _isIPv4Address(ip):
    """
    Is this a valid IPv4 address?
    """
    try:
        ipaddress.IPv4Address(ip)
        return True
    except ipaddress.AddressValueError:
        return False


class Ip(Controller):
    class Meta:
        label = 'ip'
        stacked_type = 'nested'
        stacked_on = 'base'
        help = 'Interfaces for DH IP objects (dh-ip)'
        arguments = [
            (['ipnetmac'],
             {'help': 'IP/Network/MAC Address',
              'action': 'store'}),
            (['-q', '--quiet'],
             {'help': "(eg don't print IPs belonging to mac_address)",
              'action': 'store_true'}),
            (['-s', '--servicectl'],
             {'help': '(eg "man export"): runs this servicectl on the machine owning that IP',
              'action': 'store'})
        ]

    def _fetch_ip4addr(self, address=None):
        """
        Fetch all the data we need when given an IPv4Address
        """

        address = address or self.app.pargs.ipnetmac
        details = self.app.get_url(
            self.app.config.get('ndn', 'ip_api') + '/' + address,
            headers=self.app.AUTH_HEADER)
        if not details:
            self.app.log.warning(
                "Unable to fetch IPv4 details for IP %s" % address)
            self.app.exit_code = 1
            self.app.close()

        details['networkObject'] = self.app.get_url(
            self.app.config.get('ndn', 'network_api') +
            '/' + details['network'],
            headers=self.app.AUTH_HEADER, params={'params': True})
        if not details['networkObject']:
            self.app.log.warning(
                "Unable to fetch network details for network %s" % details['network'])
            self.app.exit_code = 1
            self.app.close()

        if details['machine']:
            details['machineObject'] = self.app.get_url(
                self.app.config.get('ndn', 'machine_api') +
                '/' + details['machine'],
                headers=self.app.AUTH_HEADER)
            if not details['machineObject']:
                self.app.log.warning(
                    "Unable to fetch network details for network %s" % details['network'])
                self.app.exit_code = 1
                self.app.close()

        if self.app.get_param_from_list(details['networkObject']['params'], 'openstack_cluster'):
            details['openstack'] = self.app.get_url(
                self.app.config.get(
                    'ndn', 'openstack_port_api') + '/' + address,
                headers=self.app.AUTH_HEADER)

        return details

    @ex(help='Returns IP info for rows that match the input (synonym for "info")')
    def get(self):
        self._get_or_info()

    @ex(help='Returns IP info for rows that match the input (synonym for "get")')
    def info(self):
        self._get_or_info()

    def _get_or_info(self):
        self.app.require_token()

        # what type of query has the user provided us?
        qtype = None

        if self.app.pargs.servicectl:
            print('This tool does not yet support scheduling sctls')
            self.app.exit_code = 1
            self.app.close()
        else:
            if _isMacAddress(self.app.pargs.ipnetmac):
                self.app.log.debug('checking input for mac address type')
                result = self.app.get_url(
                    self.app.config.get('ndn', 'ip_api') +
                    '/mac/' + self.app.pargs.ipnetmac,
                    headers=self.app.AUTH_HEADER)
                if result:
                    qtype = 'mac'
            elif _isIPv4Address(self.app.pargs.ipnetmac):
                self.app.log.debug('checking input for IP type')
                result = self._fetch_ip4addr()
                if result:
                    qtype = 'ip'
            else:
                # first check is the requested input is a machine name
                self.app.log.debug('checking input for machine name type')
                result = self.app.get_url(
                    self.app.config.get('ndn', 'ip_api') +
                    '/machine/' + self.app.pargs.ipnetmac,
                    headers=self.app.AUTH_HEADER)
                if result:
                    # fetch ip4addr details, now that we know the IP
                    qtype = 'machinename'

                # if we didn't find the machine name in the `ip` table, check the `machine` table
                if not qtype:
                    self.app.log.debug('checking input for type')
                    result = self.app.get_url(
                        self.app.config.get('ndn', 'machine_api') + '/search',
                        json={'machine': self.app.pargs.ipnetmac},
                        headers=self.app.AUTH_HEADER)
                    if result:
                        qtype = 'machinename'

        # add the args to the output to assist with formatting the output
        if isinstance(result, dict):
            result['args'] = self.app.pargs.__dict__

        # figure out how to print the data we got (or didn't get)
        if qtype:
            self.app.log.debug("printing template for %s" % qtype)
            self.app.render({'data': result}, 'ip/%s.jinja2' % qtype)
        else:
            print("Unable to find any IP info for %s" %
                  self.app.pargs.ipnetmac)
