from cement import Controller, ex
from pprint import pprint
import time
import tempfile
from subprocess import call
from machine_service_client.machine import Machine as MachineClient
from machine_service_client.core import MachineServiceBase


def _cleanNullTerms(d):
    return {
        k: v
        for k, v in d.items()
        if v is not None
    }


class Machine(Controller):
    class Meta:
        label = 'machine'
        stacked_type = 'nested'
        stacked_on = 'base'
        help = 'Interfaces for DH Machine object (dh-machine)'
        arguments = [
            (['machine'],
             {'help': 'Machine name',
              'action': 'store'}),
            (['--history', '--log'],
             {'help': 'Show table of machine history',
              'action': 'store_true',
              'dest': 'log',
              'default': False}),
            (['-p', '--params'],
             {'help': 'Display all params',
              'action': 'store_true',
              'dest': 'params',
              'default': False}),
            (['-v', '--verbose'],
             {'help': 'Show more info (eg vshost vsguest list)',
              'action': 'store_true',
              'dest': 'verbose',
              'default': False}),
            (['-s', '--search'],
             {'help': 'Search for a machine (mutually exclusive with other flags)',
              'action': 'store',
              'dest': 'search',
              'default': False}),
            (['-l', '--loc', '--location'],
             {'help': 'Set location (used with Create or Modify)',
              'action': 'store',
              'dest': 'location',
              'default': False}),
            (['-t', '--type'],
             {'help': 'Set type (used with Create or Modify)',
              'action': 'store',
              'dest': 'type',
              'default': False}),
            (['-D', '--domain'],
             {'help': 'Set domain (used with Create or Modify)',
              'action': 'store',
              'dest': 'domain',
              'default': False}),
            (['-i', '--ip'],
             {'help': 'Set IP (used with Create or Modify)',
              'action': 'store',
              'dest': 'ip',
              'default': False}),
            (['--remove-ip'],
             {'help': 'Remove IP (used with Modify)',
              'action': 'store_true',
              'dest': 'remove_ip',
              'default': False}),
            (['-S', '--status'],
             {'help': 'Set status (used with Create or Modify)',
              'action': 'store_true',
              'dest': 'status',
              'default': False}),
            (['--toggle'],
             {'help': 'Toggle status (used with Modify)',
              'action': 'store_true',
              'dest': 'toggle',
              'default': False}),
            (['-u', '--user'],
             {'help': 'Return a list of machines owned by a user',
              'action': 'store',
              'dest': 'user',
              'default': False}),
            (['-L', '--list'],
             {'help': 'When searching for machines, if a single one is returned, just return the name instead of the whole machine details',
              'action': 'store_true',
              'dest': 'list',
              'default': False}),
            (['-z', '--zabbix'],
             {'help': 'Display relevant zabbix information (dashboard/alerts/etc.)',
              'action': 'store_true',
              'dest': 'zabbix',
              'default': False}),
            (['-H', '--http', '--https'],
             {'help': 'Print domains in https format.',
              'action': 'store_true',
              'dest': 'https',
              'default': False}),
        ]

    def _get_machine(self, client=None):
        client = client or MachineServiceBase(
            url=self.app.config.get('ndn', 'machine_api_host'))
        machine = MachineClient({'machine': self.app.pargs.machine}, client)
        try:
            machine.get(
                log=self.app.pargs.log,
                params=True,
                account=True,
                domainservices=True,
                extended=True)
            machines = {'logical': None, 'physical': machine}
            result = {'machine': {'logical': None,
                                  'physical': machine.machine.model_dump()}}
            if machine.machine.realm == 'logical':
                # machine is logical, lookup the physical
                machines = {'logical': machine, 'physical': None}
                if not machine.machine.vmhost or not machine.physical:
                    physical = MachineClient(
                        {'machine': machine.machine.vmhost or machine.physical}, client)
                    physical.get()
                    machines = {'logical': machine, 'physical': physical}
                    result = {'machine': {'logical': machine.machine.model_dump(
                    ), 'physical': physical.machine.model_dump()}}
        except Exception as err:
            self.app.log.error(f'Failed getting machine info: {err}')
            self.app.exit_code = 1
            self.app.close()

        for mtype in ('logical', 'physical'):
            # fetch preorder, if this is vserver/lxc
            machine = machines[mtype]
            if machine:
                result['machine'][mtype]['assigned_clusters'] = [
                    clus.cluster.cluster for clus in machine.assigned_clusters]
                result['machine'][mtype]['inherited_clusters'] = [
                    clus.cluster.cluster for clus in machine.inherited_clusters]
                if machine.isVServerGuest or machine.isLxcInstance:
                    result['machine'][mtype]['vps_preorder'] = machine.vps_preorder
                    result['machine'][mtype]['vps_version'] = machine.vps_version
                # dp domains
                result['machine'][mtype]['domains'] = self._get_dp_domains(
                    machine)
                # hasAdmin? this endpoint does not work yet, as the concept of "hasAdmin" is incredibly complex
                result['machine'][mtype]['hasAdminUser'] = False

                # param-based machine details needed for generating the output
                result['machine'][mtype]['sc_deactivated_ftp'] = machine.get_param(
                    'sc_deactivated_ftp')
                result['machine'][mtype]['u_deactivated_jabber_transports'] = machine.get_param(
                    'u_deactivated_jabber_transports')
                result['machine'][mtype]['unmanaged_service'] = machine.get_param(
                    'unmanaged_service')
                result['machine'][mtype]['u_activated_courier'] = machine.get_param(
                    'u_activated_courier')
                result['machine'][mtype]['http_port'] = machine.get_param(
                    'http_port')
                result['machine'][mtype]['phpini_cache'] = machine.get_param(
                    'phpini_cache')
                result['machine'][mtype]['servicectl_ssh_port'] = machine.get_param(
                    'servicectl_ssh_port')

                if machine.isLxcHost:
                    result['machine'][mtype]['lxc_instances'] = machine.lxc_instances()

                if machine.isOpenstack:
                    result['machine'][mtype]['openstack_info'] = machine.openstack()

        return result

    def _get_dp_domains(self, machine):
        result = []
        self.app.log.debug("Fetching domains for %s" % machine)
        domains = []
        for domain in machine.domains():
            domains.append({'domain': domain.get('domain', ''), 'po': domain.get(
                'preorder', ''), 'install': domain.get('dh_id', '')})
        for domain in [i for n, i in enumerate(domains) if i not in domains[n + 1:]]:
            result.append(domain)
        return result

    def _get_zabbix_problems(self, machine):
        result = []
        severities = ['[NOT CLASSIFIED]', '[INFO]',
                      '[WARN]', '[AVERAGE]', '[HIGH]', '[DISASTER]']
        self.app.log.debug("Fetching zabbix problems for %s" % machine)
        for problem in machine.zabbix.problems():
            result.append({
                'severity': severities[int(problem['severity'])],
                'name': problem['name'],
                'clock': time.strftime(
                    '%Y-%m-%d %H:%M:%S',
                    time.localtime(int(problem['clock'])))})
        return result

    def _get_or_info(self, client=None):
        self.app.require_token()
        client = client or MachineServiceBase(
            url=self.app.config.get('ndn', 'machine_api_host'))
        result = self._get_machine(client)
        for mtype in result['machine']:
            if result['machine'][mtype]:
                result['machine'][mtype]['zabbix'] = {}
                if self.app.pargs.zabbix:
                    self.app.log.debug(
                        "Fetching zabbix information about machine: %s" % result['machine'][mtype]['machine'])
                    result['machine'][mtype]['zabbix']['proxy'] = result['machine'][mtype].zabbix.proxy
                    result['machine'][mtype]['zabbix']['dashboard'] = result['machine'][mtype].zabbix.dashboard
                    result['machine'][mtype]['zabbix']['problems'] = result['machine'][mtype].problems
                # saving these args into the machine type, to make a easier to access during template rendering
                result['machine'][mtype]['args'] = self.app.pargs.__dict__

        self.app.render(result, 'machine/info.jinja2')

    @ex(help='Display machine information (synonym for "info")')
    def get(self):
        client = MachineServiceBase(
            url=self.app.config.get('ndn', 'machine_api_host'))
        self._get_or_info(client)

    @ex(help='Display machine information (synonym for "get")')
    def info(self):
        client = MachineServiceBase(
            url=self.app.config.get('ndn', 'machine_api_host'))
        self._get_or_info(client)

    @ex(help='Fetch Kentik URL for a machine')
    def kentik(self):
        self.app.require_token()
        print(MachineClient(url=self.app.config.get(
            'ndn', 'machine_api_host')).kentik(self.app.pargs.machine))

    @ex(help='Create new machine')
    def create(self):
        self.app.require_token()
        mclient = MachineClient(
            _cleanNullTerms({
                'machine': self.app.pargs.machine or None,
                'location': self.app.pargs.location or None,
                'domain': self.app.pargs.domain or None,
                'ip': self.app.pargs.ip or None,
                'type': self.app.pargs.type or None,
                'status': self.app.pargs.status or None,
                'comment': args.comment or None,
                'realm': args.realm or None,
                'masq': args.masq or None,
                'rack': args.rack or None,
                'nfsroot': args.nfsroot or None,
                'vmhost': args.vmhost or None}),
            url=self.app.config.get('ndn', 'machine_api_host'))

        mclient.save()
        mclient.get()
        pprint.pprint(mclient.machine.model_dump())
        return mclient.machine.model_dump()
        # for mtype in machine['machine']:
        #    if machine['machine'][mtype]:
        #        result = {'machine': {mtype: machine['machine'][mtype]}}
        #        result['machine'][mtype]['args'] = self.app.pargs.__dict__

        # self.app.render(result, 'machine/info.jinja2')

    @ex(help='Update an existing machine')
    def modify(self):
        print("This utility cannot modify/update machines yet")
        # self.app.require_token()
        # self.app.render(result, 'machine/modify.jinja2')

    @ex(help='Toggle the status of an existing machine in DH DB and in zabbix')
    def toggle(self):
        self.app.require_token()

        # check if the requested machine exists
        machine = self.app.get_url(
            self.app.config.get('ndn', 'machine_api') +
            '/%s' % self.app.pargs.machine,
            headers=self.app.AUTH_HEADER)
        if not machine:
            print("Could not find machine %s" % self.app.pargs.machine)
            self.app.exit_code = 1
            self.app.close()

        if machine['status'] == 'up':
            target_zstatus = 'disable'
            target_dstatus = 'down'
        else:
            target_zstatus = 'enable'
            target_dstatus = 'up'

        ztoggle = self.app.httpsession.get(
            self.app.config.get('ndn', 'zabbix_api') + '/' +
            self.app.pargs.machine + '/' + target_zstatus,
            headers=self.app.AUTH_HEADERS)
        if ztoggle.status_code != 200:
            print("Failed to set host %s to %s in zabbix" %
                  (self.app.pargs.machine, target_zstatus))

        dtoggle = self.app.post_json_url(
            self.app.config.get('ndn', 'machine_api'),
            json={'machine': self.app.pargs.machine, 'status': target_dstatus},
            headers=self.app.AUTH_HEADERS)
        if not dtoggle:
            print("Failed to set host %s to %s in dh-db" %
                  (self.app.pargs.machine, target_dstatus))

    @ex(help='Destroy a machine')
    def destroy(self):
        self.app.require_token()

        if input('WARNING!!! Machine %s will be DESTROYED. Are you sure (y/N)? ' % self.app.pargs.machine) == 'y':
            mdelete = self.app.delete_json_url(
                self.app.config.get('ndn', 'machine_api') +
                '/' + self.app.pargs.machine,
                headers=self.app.AUTH_HEADER)
            print(mdelete)
        else:
            print("'%s' lives to see another day" % self.app.pargs.machine)

    @ex(help='Search for a machine')
    def search(self):
        self.app.require_token()
        mclient = MachineClient({'machine': self.app.pargs.machine},
                                url=self.app.config.get('ndn', 'machine_api_host'))
        result = []
        if self.app.pargs.machine == '*':
            search_data = {}
            for attrib in [
                    "comment",
                    "domain",
                    "hardware_description",
                    "ip",
                    "location",
                    "masq",
                    "nfsroot",
                    "rack",
                    "realm",
                    "status",
                    "type",
                    "vmhost"]:
                if attrib in self.app.pargs.__dict__.keys() and self.app.pargs.__dict__.get(attrib):
                    search_data[attrib] = self.app.pargs.__dict__.get(attrib)
            result = MachineClient(search_data, url=self.app.config.get(
                'ndn', 'machine_api_host')).search()
        else:
            for machine in MachineClient(
                    {'machine': self.app.pargs.machine},
                    url=self.app.config.get('ndn', 'machine_api_host')).search(log=self.app.pargs.log):
                result.append(machine.machine)
        if result:
            print(', '.join(result))

    @ex(help='Fetch the domains associated with a machine')
    def domains(self):
        self.app.require_token()
        mclient = MachineClient({'machine': self.app.pargs.machine},
                                url=self.app.config.get('ndn', 'machine_api_host'))
        result = []

        for domain in mclient.domains():
            result.append(domain['domain'])
        for d in list(set(result)):
            if self.app.pargs.https:
                print("https://" + d)
            else:
                print(d)

    @ex(help='Launch an ssh shell on the machine')
    def connect(self):
        mclient = MachineClient({'machine': self.app.pargs.machine},
                                url=self.app.config.get('ndn', 'machine_api_host'))
        mclient.get()
        if not mclient.machine.ip:
            self.app.log.error('No IP found for machine %s' %
                               self.app.pargs.machine)
            self.app.exit_code = 1
            self.app.close()
        # write key to a temp file so ssh client can use it
        root_ssh_key = self.app.get_controller_root_ssh_key()
        temp_key = tempfile.NamedTemporaryFile(buffering=0)
        self.app.log.debug(
            'storing ssh key temporarily in file %s' % temp_key.name)
        temp_key.write(root_ssh_key.encode())

        cmd = 'ssh -l root -i %s %s' % (temp_key.name, mclient.machine.ip)
        self.app.log.debug('Running shell command: %s' % cmd)
        call(cmd, shell=True)

    @ex(help='Fetch the IPs (physical and logical) for a machine. This is an optimized IP-only lookup. Uses machine '
             'table, not IP table (which have known conflicts)',
        arguments=[
            (['--physical'],
             {'help': 'Display only the physical IP (mutually exclusive with "--logical")',
              'action': 'store_true',
              'default': False}),
            (['--logical'],
             {'help': 'Display only the logical IP (mutually exclusive with "--physical")',
              'action': 'store_true',
              'default': False}),
            (['additional_machines'],
             {'help': 'Additional machine names',
              'action': 'store',
              'default': [],
              'nargs': '*'}),
        ])
    def ip(self):
        self.app.require_token()
        result = {}
        mclient = MachineClient(
            url=self.app.config.get('ndn', 'machine_api_host'))
        # we have to concat these 2 because the base machine module has
        # "machine" as a required positional param and we are adding
        # "additional_machines" to optimize the lookup
        for machine in [self.app.pargs.machine] + self.app.pargs.additional_machines:
            result[machine] = {'physical_ip': None, 'logical_ip': None}
            mclient.get(machine)
            if not mclient.machine.ip:
                self.app.log.error('No IP found for machine %s' %
                                   self.app.pargs.machine)
                self.app.exit_code = 1
                self.app.close()

            if mclient.machine.realm == 'physical':
                result[machine]['physical_ip'] = mclient.machine.ip

            if mclient.machine.realm == 'logical':
                result[machine]['logical_ip'] = mclient.machine.ip

            # fetch logical/physical if we only have physical/logical and the user didn't specify
            if not self.app.pargs.logical and not self.app.pargs.physical:

                # if we only have the physical, get the logical
                if result[machine].get('physical_ip') and not result[machine].get('logical_ip'):
                    mclient.get(mclient.physical)
                    result[machine]['logical_ip'] = mclient.machine.ip

                # if we only have the logical, get the physical
                if result[machine].get('logical_ip') and not result[machine].get('physical_ip'):
                    mclient.get(mclient.physical)
                    result[machine]['physical_ip'] = mclient.machine.ip

            # if the user requested only the physical IP, pop off the logical IP and ensure physical is set
            if self.app.pargs.physical:
                result[machine].pop('logical_ip')
                # we don't have the physical yet. user must have passed in a logical
                if not result[machine].get('physical_ip'):
                    mclient.get(mclient.physical)
                    result[machine]['physical_ip'] = mclient.machine.ip

            # if the user requested only the logical IP, pop off the physical IP
            if self.app.pargs.logical:
                result[machine].pop('physical_ip')

        self.app.render({'result': result}, 'machine/ip.jinja2')
