from cement import Controller, ex
from servicectl_client.services import Service as ServiceAPI


class Service(Controller):
    class Meta:
        label = 'service'
        stacked_type = 'nested'
        stacked_on = 'base'
        help = 'Get information about and create new DH services'

    @ex(
        help='Fetch a list of registered services',
        arguments=[
            (['-S', '--service'],
             {'help': 'Service name',
              'default': None,
              'action': 'store'}),
            (['-H', '--home'],
             {'help': 'Service home',
              'default': None,
              'action': 'store'}),
            (['-s', '--status'],
             {'help': 'Service status',
              'default': None,
              'action': 'store'}),
            (['-t', '--type'],
             {'help': 'Service type',
              'default': None,
              'action': 'store'}),
            (['-p', '--parent_home'],
             {'help': 'Parent home',
              'default': None,
              'action': 'store'}),
            (['-P', '--parent_service'],
             {'help': 'Parent service',
              'default': None,
              'action': 'store'}),
        ])
    def list(self):
        self.app.require_token()
        services = ServiceAPI(service=self.app.pargs.__dict__).list()
        if services:
            self.app.render({'data': {'args': self.app.pargs.__dict__,
                            'services': services}}, 'service/list.jinja2')
        else:
            self.app.log.error('Failed to find any services matching criteria')
            self.app.exit_code = 1
            self.app.close()

    @ex(
        help='Fetch details about a specific service',
        arguments=[
            (['service'],
             {'help': 'Service name in the form of "home:service"',
              'default': None,
              'action': 'store'}),
            (['-p', '--params'],
             {'help': 'Include service params in output',
              'default': False,
              'action': 'store_true'}),
        ])
    def get(self):
        self.app.require_token()
        home, service = self.app.pargs.service.split(':')
        result = ServiceAPI(service={'home': home, 'service': service}).get(
            params=self.app.pargs.params)
        if not result:
            self.app.log.error(
                'Unable to locate service %s:%s' % (home, service))
            self.app.exit_code = 1
            self.app.close()
        else:
            self.app.render({'data': {'args': self.app.pargs.__dict__},
                            'service': result.service.model_dump()}, 'service/details.jinja2')

    @ex(
        help='Register a new service',
        arguments=[
            (['service'],
             {'help': 'Service to create in the form of "home:service"',
              'default': None,
              'action': 'store'}),
            (['-f', '--force'],
             {'help': 'Force the creation of this service, bypassing validation checks',
              'default': False,
              'action': 'store_true'}),
            (['-p', '--parent'],
             {'help': 'Parent service in the form of "home:service"',
              'default': None,
              'action': 'store'}),
            (['-c', '--controller'],
             {'help': 'Controller service in the form of "home:service"',
              'default': None,
              'action': 'store'}),
            (['-t', '--type'],
             {'help': 'Service type',
              'default': None,
              'action': 'store'}),
            (['-s', '--status'],
             {'help': 'Service status (active/inactive)',
              'default': 'active',
              'action': 'store'}),
        ])
    def create(self):
        self.app.require_token()
        home, service = self.app.pargs.service.split(':')
        if not service:
            self.app.log.error(
                'Please pass service in the form of "home:service"')
            self.app.exit_code = 1
            self.app.close()
        phome, pservice = self.app.pargs.parent.split(
            ':') if self.app.pargs.parent else (None, None)
        chome, cservice = self.app.pargs.controller.split(
            ':') if self.app.pargs.controller else (None, None)

        result = self.app.post_json_url(
            self.app.config.get('ndn', 'servicectl_api') + '/services',
            headers=self.app.AUTH_HEADER,
            json={
                'home': home,
                'service': service,
                'status': self.app.pargs.status,
                'type': self.app.pargs.type,
                'parent_home': phome,
                'parent_service': pservice,
                'controller_home': chome,
                'controller_service': cservice},
            params={'force': self.app.pargs.force},
            expected_status=201)
        if not result:
            self.app.log.error(
                'Failed to create service %s:%s. Re-run with debug (-d) to get more details about the failure.', (home, service))
            self.app.exit_code = 1
            self.app.close()
        else:
            self.app.render({'data': {'args': self.app.pargs.__dict__},
                            'service': result}, 'service/details.jinja2')

    @ex(
        help='Delete a service',
        arguments=[
            (['service'],
             {'help': 'Service to delete in the form of "home:service"',
              'default': None,
              'action': 'store'}),
            (['-f', '--force'],
             {'help': 'Force the deletion of this service, bypassing validation checks',
              'default': False,
              'action': 'store_true'}),
            (['-p', '--skip_params'],
             {'help': 'Skip deleting parameters of this service',
              'default': False,
              'action': 'store_true'}),
        ])
    def delete(self):
        self.app.require_token()
        home, service = self.app.pargs.service.split(':')
        if not service:
            self.app.log.error(
                'Please pass service in the form of "home:service"')
            self.app.exit_code = 1
            self.app.close()

        result = self.app.httpsession.delete(
            self.app.config.get('ndn', 'servicectl_api') +
            '/services/%s/%s' % (home, service),
            headers=self.app.AUTH_HEADER,
            params={'force': self.app.pargs.force, 'skip_params': self.app.pargs.skip_params})
        if result.ok:
            print(result.json())
        else:
            self.app.log.error(result.json()['detail'])
            self.app.exit_code = 1
            self.app.close()

    @ex(
        help='Search for cron-like services. Must pass only one of "-H", "-c", or "-t"',
        arguments=[
            (['-H', '--home'],
             {'help': 'Search for a service based on the home (machine or cluster)',
              'default': None,
              'action': 'store'}),
            (['-c', '--channel'],
             {'help': 'Limit search to a specific channel',
              'default': None,
              'action': 'store'}),
            (['-t', '--type'],
             {'help': 'Limit search to a specific service type',
              'default': None,
              'dest': 'stype',
              'action': 'store'})
        ])
    def crond(self):
        self.app.require_token()
        if self.app.pargs.home:
            params = {'home': self.app.pargs.home}
        elif self.app.pargs.channel:
            params = {'channel': self.app.pargs.channel}
        elif self.app.pargs.stype:
            params = {'stype': self.app.pargs.stype}
        else:
            self.app.log.error('Please pass only one of "-H", "-c" or "-t".')
            self.app.exit_code = 1
            self.app.close()

        result = self.app.get_url(
            self.app.config.get('ndn', 'servicectl_api') + '/services/crons',
            headers=self.app.AUTH_HEADER,
            params=params)
        if not result:
            self.app.log.error(
                'Unable to locate any cron-like services for your search')
            self.app.exit_code = 1
            self.app.close()
        else:
            self.app.render({'data': {'args': self.app.pargs.__dict__},
                            'results': result}, 'service/cron_like.jinja2')
