from cement import Controller, ex
from getpass import getuser
from datetime import datetime, timedelta
import json
from servicectl_client.jobs import Job
from servicectl_client.services import Service


SCARGS = [
    'service',
    'home',
    'options',
    'kill_after',
    'kill_interval',
    'channel',
    'priority',
    'instance',
    'hinstance',
    'asap',
    'location_hint',
]


class ServiceCtl(Controller):
    class Meta:
        label = 'servicectl'
        stacked_type = 'nested'
        stacked_on = 'base'
        help = 'CLI scheduling of Service Control jobs'
        arguments = [
            (['-v', '--verbose'],
             {'help': 'Increase output verbosity',
              'action': 'store_true'})]

    @ex(help='Schedule a Service Control job',
        arguments=[
            (['-s', '--service'],
             {'help': 'Service name',
              'action': 'store'}),
            (['-H', '--home'],
             {'help': 'Service home',
              'action': 'store'}),
            (['-c', '--command'],
             {'help': 'Service command',
              'dest': 'sc_command',
              'action': 'store'}),
            (['-o', '--options'],
             {'help': 'Service options',
              'action': 'store'}),
            (['-k', '--kill-after'],
             {'help': 'Kill the service after X seconds',
              'dest': 'kill_after',
              'action': 'store'}),
            (['-K', '--kill-interval'],
             {'help': 'How frequently to send the kill command',
              'dest': 'kill_interval',
              'action': 'store'}),
            (['-C', '--channel'],
             {'help': 'Channel to schedule Service control in',
              'default': 'default',
              'action': 'store'}),
            (['-p', '--priority'],
             {'help': 'Job priority',
              'default': 0,
              'action': 'store'}),
            (['-i', '--instance'],
             {'help': 'Instance',
              'action': 'store'}),
            (['-I', '--hinstance'],
             {'help': 'Home instance',
              'action': 'store'}),
            (['-a', '--asap'],
             {'help': 'As soon as possible',
              'default': 0,
              'action': 'store'}),
            (['-l', '--location-hint'],
             {'help': 'Location hint',
              'dest': 'location_hint',
              'action': 'store'}),
            (['-S', '--scheduled'],
             {'help': 'Scheduled time to execute the job (not functional currently)',
              'default': datetime.now(),
              'action': 'store'}),
        ])
    def schedule(self):
        # get list of non-null args
        pargs = self.app.pargs.__dict__
        params = {arg: pargs.get(arg) for arg in pargs if pargs.get(
            arg, False) and arg in SCARGS}
        # add in the `sc_command` arg, because `command` is already used in python-cement, so we needed a different name
        if pargs.get('sc_command', False):
            params['command'] = pargs.get('sc_command')

        # add the current user as "executed_by_user"
        params['executed_by_user'] = getuser()

        # check for required args
        for req_arg in ['service', 'home', 'command']:
            if req_arg not in params:
                self.app.log.error('Missing required parameter "%s"' % req_arg)
                self.app.exit_code = 1
                self.app.close()

        # send sc schedule request to the API
        job = Job(params)
        schedule = job.schedule().dict()

        # add CLI args to the output
        schedule['args'] = self.app.pargs.__dict__
        self.app.render({'data': schedule}, 'servicectl/job.jinja2')

    @ex(
        help='Fetch information about a scheduled servicectl (by ID)',
        arguments=[
            (['sqid'],
             {'help': 'Servicectl job ID',
              'default': None,
              'action': 'store'}),
        ])
    def get(self):
        job = Job({'service_queue_id': self.app.pargs.sqid})
        job_status = job.get().dict()
        if not job_status:
            self.app.log.error(
                'Failed to get job status for SQID "%s" (re-run with -d for more details)' % self.app.pargs.sqid)
            self.app.exit_code = 1
            self.app.close()
        job_status['args'] = self.app.pargs.__dict__
        self.app.render({'data': job_status}, 'servicectl/job.jinja2')

    @ex(
        help='Fetch logs for a servicectl job (by ID)',
        arguments=[
            (['sctl_id'],
             {'help': 'Servicectl job ID',
              'default': None,
              'action': 'store'}),
        ])
    def logs(self):
        job = Job({'service_queue_id': self.app.pargs.sctl_id})
        job_logs = job.logs()
        if job_logs:
            print(''.join(job_logs))
        else:
            self.app.log.error(
                'Failed to find logs for SQID "%s"' % self.app.pargs.sctl_id)
            self.app.exit_code = 1
            self.app.close()

    @ex(
        help='Search the service queue for scheduled jobs',
        arguments=[
            (['-s', '--service'],
             {'help': 'Service name',
              'action': 'store',
              'default': None}),
            (['-H', '--home'],
             {'help': 'Service home',
              'action': 'store',
              'default': None}),
            (['-c', '--command'],
             {'help': 'Service command',
              'dest': 'sc_command',
              'default': None}),
            (['-i', '--interval'],
             {'help': 'Time interval to bind the search to. Must be in the format 5d, 12h, or 42m. (default: 1d)',
              'action': 'store',
              'default': '1d'}),
            (['-C', '--channel'],
             {'help': 'Job channel',
              'action': 'store',
              'default': None}),
            (['-f', '--finished'],
             {'help': 'Field tracking whether the job has finished. Use "null" if you want to search for jobs that have not yet finished',
              'action': 'store',
              'default': None}),
            (['-r', '--result'],
             {'help': 'Field tracking the job result.',
              'action': 'store',
              'default': None}),
            (['--sqid_only'],
             {'help': 'Return just the IDs, space separated (great for piping to other scripts).',
              'action': 'store_true',
              'default': False}),
        ]
    )
    def search(self):
        search = {}
        search_args = {
            'service': self.app.pargs.service,
            'home': self.app.pargs.home,
            'command': self.app.pargs.sc_command,
            'channel': self.app.pargs.channel,
            'result': self.app.pargs.result}
        for sarg in search_args:
            if sarg == 'finished':
                continue
            if search_args[sarg]:
                search[sarg] = search_args[sarg]
        if self.app.pargs.finished:
            search['finished'] = None if self.app.pargs.finished == 'null' else self.app.pargs.finished

        # translate interval to minutes to simplify the timedelta calculation
        if self.app.pargs.interval.lower()[-1] == 'd':
            interval_minutes = int(
                self.app.pargs.interval.lower()[:-1]) * 24 * 60
        elif self.app.pargs.interval.lower()[-1] == 'h':
            interval_minutes = int(self.app.pargs.interval.lower()[:-1]) * 60
        elif self.app.pargs.interval.lower()[-1] == 'm':
            interval_minutes = int(self.app.pargs.interval.lower()[:-1])
        else:
            self.app.log.error(
                'Unable to determine interval. Must be in the format 5d, 12h, or 42m')
            self.app.exit_code = 1
            self.app.close()
        search['scheduled'] = datetime.now(
        ) - timedelta(minutes=interval_minutes)

        # translate to json string, then back into py dict so we're certain we're in a good
        # format before we pass it to the API endpoint
        job = Job(search)
        results = job.search()
        if results:
            if self.app.pargs.sqid_only:
                print(' '.join([str(i['service_queue_id']) for i in results]))
                self.app.exit_code = 0
                self.app.close()
            else:
                self.app.render(
                    {'data': results,
                     'command_len': max([len(result.command) for result in results])},
                    'servicectl/job_search.jinja2')
        else:
            self.app.log.error(
                'Failed to get job search results from API. Rerun with -d for more info')
            self.app.exit_code = 1
            self.app.close()

    @ex(
        help='Reap a servicectl job, marking it as "died" and filling in the proper fields',
        arguments=[
            (['sctl_id'],
             {'help': 'Servicectl job ID',
              'default': None,
              'action': 'store'}),
        ])
    def reap(self):
        job_status = self.app.get_url(
            self.app.config.get('ndn', 'servicectl_api') +
            '/jobs/reap',
            headers=self.app.AUTH_HEADER)
        if job_status:
            print("Successfully reaped SQID %s" %
                  job_status['service_queue_id'])
        else:
            self.app.log.error('Failed to reap SQID "%s"' %
                               self.app.pargs.sctl_id)
            self.app.exit_code = 1
            self.app.close()

    @ex(
        help='Cancel a servicectl job',
        arguments=[
            (['sctl_id'],
             {'help': 'Servicectl job ID',
              'default': None,
              'action': 'store',
              'nargs': '*'}),
        ])
    def cancel(self):
        for sqid in self.app.pargs.sctl_id:
            job = Job({'service_queue_id': sqid})
            if job.cancel():
                print("Successfully canceled SQID %s" % sqid)
            else:
                self.app.log.error('Failed to cancel SQID "%s"' % sqid)
                self.app.exit_code = 1
                self.app.close()

    @ex(
        help='Set a servicectl job to "died"',
        arguments=[
            (['sctl_id'],
             {'help': 'Servicectl job ID',
              'default': None,
              'action': 'store',
              'nargs': '*'}),
        ])
    def died(self):
        for sqid in self.app.pargs.sctl_id:
            job = Job({'service_queue_id': sqid})
            if job.died():
                print('Successfully set SQID %s to "died"' % sqid)
            else:
                self.app.log.error('Failed to set SQID "%s" to "died"' % sqid)
                self.app.exit_code = 1
                self.app.close()

    @ex(
        help='Set a servicectl job to "finished"',
        arguments=[
            (['sctl_id'],
             {'help': 'Servicectl job ID',
              'default': None,
              'action': 'store',
              'nargs': '*'}),
        ])
    def finish(self):
        for sqid in self.app.pargs.sctl_id:
            job = Job({'service_queue_id': sqid})
            if job.finish():
                print('Successfully set SQID %s to "finished"' % sqid)
            else:
                self.app.log.error(
                    'Failed to set SQID "%s" to "finished"' % sqid)
                self.app.exit_code = 1
                self.app.close()

    @ex(
        help='Kill a servicectl job',
        arguments=[
            (['sctl_id'],
             {'help': 'Servicectl job ID',
              'default': None,
              'action': 'store',
              'nargs': '*'}),
        ])
    def kill(self):
        for sqid in self.app.pargs.sctl_id:
            job = Job({'service_queue_id': sqid})
            if job.kill():
                print("Successfully killed SQID %s" % sqid)
            else:
                self.app.log.error('Failed to kill SQID "%s"' % sqid)
                self.app.exit_code = 1
                self.app.close()

    @ex(
        help='Fail a servicectl job',
        arguments=[
            (['sctl_id'],
             {'help': 'Servicectl job ID',
              'default': None,
              'action': 'store',
              'nargs': '*'}),
        ])
    def fail(self):
        for sqid in self.app.pargs.sctl_id:
            job = Job({'service_queue_id': sqid})
            if job.fail():
                print("Successfully failed SQID %s" % sqid)
        else:
            self.app.log.error('Failed to set SQID "%s" to "failed"' % sqid)
            self.app.exit_code = 1
            self.app.close()

    @ex(
        help='Mark a job as "execute" (fill in the executed field)',
        arguments=[
            (['sctl_id'],
             {'help': 'Servicectl job ID',
              'default': None,
              'action': 'store',
              'nargs': '*'}),
        ])
    def execute(self):
        sqids = [int(sqid) for sqid in self.app.pargs.sctl_id]
        if Job().execute(sqids=sqids):
            self.app.log.info(
                "Successfully set SQID(s) %s to executed" % sqids)
        else:
            self.app.log.error(
                'Failed to set SQIDs "%s" to "executed"' % sqids)
            self.app.exit_code = 1
            self.app.close()

    @ex(
        help='Find long-running jobs',
        arguments=[
            (['-D', '--days'],
             {'help': 'Search for jobs running longer than X days (default: 14)',
              'default': 14,
              'action': 'store'})
        ])
    def long_running_jobs(self):
        long_jobs = Job().long_running_jobs(days=self.app.pargs.days)

        if not len(long_jobs):
            self.app.log.info("No long running jobs found! \\0/")
            self.app.close()
        else:
            self.app.log.info(
                f"Found {len(long_jobs)} jobs executing more than {self.app.pargs.days} days")

        user_input = input("(F)inish, (L)ist, (E)xit? ")
        if user_input.upper() == "F":
            self.app.log.debug('User selected "finish" option...')
            self.app.log.info("finishing jobs...")
            for lj in long_jobs:
                self.app.log.info(f"  {lj.job.service_queue_id}")
                lj.finish()
            self.app.log.info("Done!")
        elif user_input.upper() == "L":
            self.app.log.debug('User selected "list" option...')
            print('\n'.join([json.dumps(json.loads(lj.job.json()))
                  for lj in long_jobs]))
        else:
            self.app.log.debug('User did not select "D" or "L" options...')
            self.app.close()

    @ex(help='List all cron-like sctl services')
    def crons(self):
        self.app.render(
            {'data': {'args': self.app.pargs.__dict__,
                      'parameters': Service().all_crons()}},
            'servicectl/cron_params.jinja2')
