from cement import Controller, ex


class ServiceParam(Controller):
    class Meta:
        label = 'param'
        stacked_type = 'nested'
        stacked_on = 'service'
        help = 'Get and set service parameters'

    @ex(
        help='List the parameters of a service',
        arguments=[
            (['service'],
             {'help': 'Service to fetch parameters for (i.e. home:service)',
              'default': None,
              'action': 'store'}),
        ])
    def list(self):
        try:
            home, service = self.app.pargs.service.split(':')
        except ValueError:
            self.app.log.error(
                'Please pass service in the form of "home:service"')
            self.app.exit_code = 1
            self.app.close()

        result = self.app.get_url(
            self.app.config.get('ndn', 'servicectl_api') +
            '/services/params/%s/%s' % (home, service),
            headers=self.app.AUTH_HEADER)
        if result:
            self.app.render({'data': {'args': self.app.pargs.__dict__,
                            'parameters': result}}, 'service/parameters.jinja2')
        else:
            self.app.log.error(
                'Failed to fetch params for service %s:%s. Re-run with debug (-d) for more details.' % (home, service))
            self.app.exit_code = 1
            self.app.close()

    @ex(
        help='Delete a parameter. Because services can have duplicates of the same param with different values, you must supply both the '
             'parameter name and matching value you wish to delete',
        arguments=[
            (['service'],
             {'help': 'Service to fetch parameters for (i.e. home:service)',
              'default': None,
              'action': 'store'}),
            (['-p', '--parameter'],
             {'help': 'Parameter name',
              'default': None,
              'action': 'store'}),
            (['-v', '--value'],
             {'help': 'Parameter value',
              'default': None,
              'action': 'store'}),
            (['-t', '--text-value'],
             {'help': 'Parameter text-value',
              'default': None,
              'action': 'store'}),
        ])
    def delete(self):
        try:
            home, service = self.app.pargs.service.split(':')
        except ValueError:
            self.app.log.error(
                'Please pass service in the form of "home:service"')
            self.app.exit_code = 1
            self.app.close()

        if not self.app.pargs.parameter and self.app.pargs.value:
            self.app.log.error('Please provide the parameter name and value')
            self.app.exit_code = 1
            self.app.close()

        param_details = {
            'home': home,
            'service': service,
            'parameter': self.app.pargs.parameter,
            'value': self.app.pargs.value,
            'text_value': self.app.pargs.text_value
        }

        result = self.app.delete_json_url(
            self.app.config.get('ndn', 'servicectl_api') + '/services/param/%s/%s/%s/%s' % (
                home,
                service,
                self.app.pargs.parameter,
                self.app.pargs.value),
            headers=self.app.AUTH_HEADER,
            json=param_details)
        if result:
            print("Successfully deleted parameter %s=%s from service %s" % (
                self.app.pargs.parameter, self.app.pargs.value, self.app.pargs.service))
        else:
            self.app.log.error('Failed to delete param %s=%s for service %s. Re-run with debug (-d) for more details.' %
                               (self.app.pargs.parameter, self.app.pargs.value, self.app.pargs.service))
            self.app.exit_code = 1
            self.app.close()

    @ex(
        help='Add a parameter to a service',
        arguments=[
            (['service'],
             {'help': 'Service to fetch parameters for (i.e. home:service)',
              'default': None,
              'action': 'store'}),
            (['-p', '--parameter'],
             {'help': 'Parameter name',
              'default': None,
              'action': 'store'}),
            (['-v', '--value'],
             {'help': 'Parameter value',
              'default': None,
              'action': 'store'}),
            (['-t', '--text-value'],
             {'help': 'Parameter text-value',
              'default': None,
              'action': 'store'}),
        ])
    def add(self):
        try:
            home, service = self.app.pargs.service.split(':')
        except ValueError:
            self.app.log.error(
                'Please pass service in the form of "home:service"')
            self.app.exit_code = 1
            self.app.close()

        if not self.app.pargs.parameter and self.app.pargs.value:
            self.app.log.error('Please provide the parameter name and value')
            self.app.exit_code = 1
            self.app.close()

        param_details = {
            'home': home,
            'service': service,
            'parameter': self.app.pargs.parameter,
            'value': self.app.pargs.value,
            'text_value': self.app.pargs.text_value
        }

        result = self.app.post_json_url(
            self.app.config.get('ndn', 'servicectl_api') + '/services/param',
            headers=self.app.AUTH_HEADER,
            json=param_details,
            expected_status=201)
        if result:
            print("Successfully added parameter %s=%s to service %s" % (
                self.app.pargs.parameter, self.app.pargs.value, self.app.pargs.service))
        else:
            self.app.log.error('Failed to add param %s=%s to service %s. Re-run with debug (-d) for more details.' %
                               (self.app.pargs.parameter, self.app.pargs.value, self.app.pargs.service))
            self.app.exit_code = 1
            self.app.close()
