
from cement import App, TestApp, init_defaults
from cement.core.exc import CaughtSignal
from .core.exc import ndnError
from .controllers.base import Base
from .controllers.machine import Machine
from .controllers.cluster import Cluster
from .controllers.ip import Ip
from .controllers.lighthouse import Lighthouse
from .controllers.ansiblebase import Ansible
from .controllers.ansibleinventory import AnsibleInventory
from .controllers.ansibleutils import AnsibleUtils
from .controllers.ansibleawx import AnsibleAWX
#from .controllers.domain import Domain
from .controllers.servicectl import ServiceCtl
from .controllers.servicectlchannels import ServiceCtlChannels
from .controllers.servicectlhealth import ServiceCtlHealth
from .controllers.vault import Vault
from .controllers.db import Db
from .controllers.service import Service
from .controllers.serviceparam import ServiceParam
from .controllers.login import Login
from .controllers.logout import Logout
from .controllers.wordpress import Wordpress
from .controllers.dhq import Dhq
from .controllers.gitlab import Gitlab, GitlabPersonal, GitlabPersonalToken
from .controllers.apis import Apis
from .controllers.version import Version
from .controllers.plugin import Plugin
from .controllers.alertsilence import AlertSilence
from os.path import exists, expanduser
from requests import Session
import hvac
import os
from getpass import getpass
from cryptography.fernet import Fernet
import cryptography
import pkg_resources
from packaging.version import Version as pversion
import time
import json

# configuration defaults
CONFIG = init_defaults('ndn')
CONFIG['ndn']['fernet_key'] = 'aUmK_pYT0lT0Kj5hCSPUZSi8mtipEASmRsiWnBR9ia8='
CONFIG['ndn']['machine_api_host'] = 'https://machine-service-master.prod-bm.kubecluster.vk.dreamhost.io'
CONFIG['ndn']['auth_api'] = CONFIG['ndn']['machine_api_host']
CONFIG['ndn']['machine_api'] = CONFIG['ndn']['machine_api_host'] + '/machine'
CONFIG['ndn']['cluster_api'] = None
CONFIG['ndn']['zabbix_api'] = CONFIG['ndn']['machine_api_host'] + '/zabbix'
CONFIG['ndn']['ip_api'] = CONFIG['ndn']['machine_api_host'] + '/ip'
CONFIG['ndn']['network_api'] = CONFIG['ndn']['machine_api_host'] + '/network'
CONFIG['ndn']['openstack_port_api'] = CONFIG['ndn']['machine_api_host'] + \
    '/openstack_port'
CONFIG['ndn']['lighthouse_api'] = CONFIG['ndn']['machine_api_host'] + '/lighthouse'
CONFIG['ndn']['ansible_awx_api'] = {
    'prod': 'https://awx.dh-int.com', 'dev': 'https://awx-dev.dev-bm.kubecluster.dreamhost.io'}
CONFIG['ndn']['domain_api_host'] = 'https://domain-service-master.dh-prod-eks.kubecluster.dreamhost.io'
CONFIG['ndn']['domain_api'] = CONFIG['ndn']['domain_api_host'] + '/domain'
CONFIG['ndn']['servicectl_api_host'] = 'https://servicectl-api-master.prod-bm.kubecluster.vk.dreamhost.io'
CONFIG['ndn']['servicectl_api'] = CONFIG['ndn']['servicectl_api_host'] + '/servicectl'
CONFIG['ndn']['vault_url'] = 'https://active-vault.dh-int.com:8200'
CONFIG['ndn']['wordpress_api_host'] = 'https://machine-service-master.prod-bm.kubecluster.vk.dreamhost.io'
CONFIG['ndn']['wordpress_api'] = CONFIG['ndn']['wordpress_api_host'] + '/wordpress'
CONFIG['ndn']['gitlab_api_host'] = 'https://git.dreamhost.com'
CONFIG['ndn']['gitlab_api'] = CONFIG['ndn']['gitlab_api_host'] + '/api/v4'
CONFIG['ndn']['backstage_api_host'] = 'https://portal.idp.dh-int.com'
CONFIG['ndn']['backstage_api'] = CONFIG['ndn']['backstage_api_host'] + '/api'
CONFIG['ndn']['alert_silence_api_host'] = 'https://prometheus-scrape-service-master.vis-bm.kubecluster.dreamhost.io'
CONFIG['ndn']['alert_silence_api'] = CONFIG['ndn']['alert_silence_api_host'] + \
    '/alert/silences'


def adjust_output_handler(app):
    # allow for controllers to provide "output_handler" in the Meta to override the default renderer for the controller
    for coh in [thing for thing in app.Meta.handlers if thing.Meta.label == app.pargs.__controller_namespace__]:
        if getattr(coh.Meta, 'output_handler', None):
            app.log.debug(
                f'Controller "{coh.Meta.label}" is overriding output_handler with "{coh.Meta.output_handler}"')
            app._meta.output_handler = coh.Meta.output_handler
            app._setup_output_handler()

    # if the --output-handler CLI flag was passed, use that as an override/force
    if app.pargs.output_handler and getattr(app._meta, 'output_handler') != app.pargs.output_handler:
        app.log.debug(
            f"Switching output handler to {app.pargs.output_handler}")
        app._meta.output_handler = app.pargs.output_handler
        app._setup_output_handler()


class ndn(App):
    """ndn primary application."""

    class Meta:
        label = 'ndn'

        # configuration defaults
        config_defaults = CONFIG

        # configuration dirs
        config_dirs = ['config/']

        # call sys.exit() on close
        exit_on_close = True

        # load additional framework extensions
        extensions = [
            'colorlog',
            'jinja2',
            'json',
            'plugin',
            'tabulate',
            'yaml',
        ]

        # configuration handler
        config_handler = 'yaml'

        # configuration file suffix
        config_file_suffix = '.yml'

        # set the log handler
        log_handler = 'colorlog'

        # set the output handler
        output_handler = 'jinja2'

        # register handlers
        # keep this in reverse sorted order so they are in alphabetical order
        # when you run `ndn --help`
        handlers = [
            Wordpress,
            Version,
            Vault,
            ServiceParam,
            ServiceCtlHealth,
            ServiceCtlChannels,
            ServiceCtl,
            Service,
            Plugin,
            Machine,
            Logout,
            Login,
            Lighthouse,
            Ip,
            GitlabPersonalToken,
            GitlabPersonal,
            Gitlab,
            # Domain,
            Dhq,
            Db,
            Cluster,
            Base,
            Apis,
            AnsibleUtils,
            AnsibleInventory,
            AnsibleAWX,
            Ansible,
            AlertSilence,
        ]
        hooks = [('post_argument_parsing', adjust_output_handler)]

    def __init__(self):
        super().__init__()
        self.creds_cache_file = expanduser('~') + '/.ndn_creds_cache'
        self.vault_token_file = expanduser('~') + '/.vault-token'
        self.jwt_token_file = expanduser('~') + '/.ndnauth'

    def require_token(self):
        if not hasattr(self, 'api_token'):
            self.log.error("API token required, but not found")
            self.exit_code = 1
            self.close()
        self.validate_token()

    def require_creds(self, username=None, password=None):
        # fetch cached creds
        (self.username, self.password) = self.get_cached_creds()
        # accept overrides, or fallback to cached creds
        self.username = username or self.username
        self.password = password or self.password
        # if we're still falsy, prompt
        if not self.username or not self.password:
            self.log.warning(
                "Username and password credentials required, but not found")
            self.prompt_ldap_login(
                getattr(self, 'username', username),
                getattr(self, 'password', password))
        else:
            self.save_cached_creds(self.username, self.password)

    def validate_token(self):
        self.log.debug("Sending auth validation to %s" %
                       self.config.get('ndn', 'auth_api') + '/auth')
        auth_resp = self.post_json_url(
            self.config.get('ndn', 'auth_api') + '/auth',
            headers={"Content-type": "application/json"},
            json={"access_token": self.api_token})

    def get_param_from_list(self, paramlist, param):
        """
        Given a list of parameters, fetch the requested param
        """
        result = [x for x in paramlist if x['parameter'] == param]
        if len(result) == 1:
            return result[0]
        elif len(result) == 0:
            return
        else:
            return result

    def get_vault_client(self):
        self.log.debug("Checking if vault client token is valid")
        client = hvac.Client(url=self.config.get('ndn', 'vault_url'))
        if os.path.isfile(self.vault_token_file):
            with open(self.vault_token_file, 'r') as vtoken:
                client.token = vtoken.read()
        try:
            if client.is_authenticated():
                self.log.debug(
                    "Successfully authed vault client with existing token at %s" % self.vault_token_file)
            else:
                self.log.warning(
                    'Vault client is not authenticated. Try running `ndn vault login`.')
        except ValueError:
            self.log.warning(
                "Failed to access vault with token at %s" % self.vault_token_file)
            self.log.warning(
                "Please login to vault to get a valid token (`ndn vault login` or using the vault client)")
        return client

    def get_controller_root_ssh_key(self):
        self.log.debug("Fetching root ssh key from vault")
        vclient = self.get_vault_client()
        if not vclient.is_authenticated():
            self.exit_code = 1
            self.close()
        key = vclient.secrets.kv.v1.read_secret(
            path='auth/ssh/controller.key', mount_point='secret')['data']['content']
        self.log.debug("Found the following controller ssh key: %s" % key)
        return key

    def get_url(self, url, params=None, headers=None, expected_status=200, result=None):
        """
        GET data from a URL and provide standard logging for non-expected results
        """
        self.log.debug("GET-ing data from URL %s" % url)
        self.log.debug("params: %s" % params)
        self.log.debug("headers: %s" % headers)
        resp = self.httpsession.get(
            url,
            headers=headers,
            params=params)
        if resp.status_code == expected_status:
            result = resp.json()
        elif resp.status_code == 403:
            self.log.warning("Permission denied for %s" % url)
            self.log.debug("response: %s" % resp.content)
            self.log.error(
                'you may need to run `ndn login` to get a fresh token')
            self.exit_code = 403
            self.close()
        else:
            self.log.debug("Failed to GET data from URL %s" % url)
            self.log.debug("status code: %s" % resp.status_code)
            self.log.debug("response: %s" % resp.content)
        return result

    def post_json_url(self, url, params=None, headers=None, expected_status=200, json={}, result=None):
        """
        POST JSON data to a URL and provide standard logging for non-expected results
        """
        self.log.debug("POST-ing data from URL %s" % url)
        self.log.debug("params: %s" % params)
        self.log.debug("headers: %s" % headers)
        self.log.debug("json: %s" % json)
        resp = self.httpsession.post(
            url,
            headers=headers,
            params=params,
            json=json)
        if resp.status_code == expected_status:
            result = resp.json()
        elif resp.status_code == 403:
            self.log.warning("Permission denied for %s" % url)
            self.log.debug("response: %s" % resp.content)
            self.log.error(
                'you may need to run `ndn login` to get a fresh token')
            self.exit_code = 403
            self.close()
        else:
            result = False
            self.log.debug("Failed to POST data to URL %s" % url)
            self.log.debug("expected status code: %s" % expected_status)
            self.log.debug("actual status code: %s" % resp.status_code)
            self.log.debug("response: %s" % resp.content)
        return result

    def put_json_url(self, url, params=None, headers=None, expected_status=201, json={}, result=None):
        """
        PUT JSON data to a URL and provide standard logging for non-expected results
        """
        self.log.debug("PUT-ing data from URL %s" % url)
        self.log.debug("params: %s" % params)
        self.log.debug("headers: %s" % headers)
        resp = self.httpsession.put(
            url,
            headers=headers,
            params=params,
            json=json)
        if resp.status_code == expected_status:
            result = resp.json()
        elif resp.status_code == 403:
            self.log.warning("Permission denied for %s" % url)
            self.log.debug("response: %s" % resp.content)
            self.log.error(
                'you may need to run `ndn login` to get a fresh token')
            self.exit_code = 403
            self.close()
        else:
            result = False
            self.log.debug("Failed to PUT data to URL %s" % url)
            self.log.debug("expected status code: %s" % expected_status)
            self.log.debug("actual status code: %s" % resp.status_code)
            self.log.debug("response: %s" % resp.content)
        return result

    def delete_json_url(self, url, params=None, headers=None, expected_status=200, json={}, result=None):
        """
        DELETE JSON data from a URL and provide standard logging for non-expected results
        """
        self.log.debug("DELETE-ing data from URL %s" % url)
        self.log.debug("params: %s" % params)
        self.log.debug("headers: %s" % headers)
        resp = self.httpsession.delete(
            url,
            headers=headers,
            params=params,
            json=json)
        if resp.status_code == expected_status:
            result = resp.json()
        elif resp.status_code == 403:
            self.log.warning("Permission denied for %s" % url)
            self.log.debug("response: %s" % resp.content)
            self.log.error(
                'you may need to run `ndn login` to get a fresh token')
            self.exit_code = 403
            self.close()
        else:
            result = False
            self.log.debug("Failed to DELETE data from URL %s" % url)
            self.log.debug("expected status code: %s" % expected_status)
            self.log.debug("actual status code: %s" % resp.status_code)
            self.log.debug("response: %s" % resp.content)
        return result

    def get_cached_creds(self):
        """
        Retrieve the cached ndn credentials
        """
        self.log.debug(
            "Attempting to retrieve credentials from ~/.ndn_creds_cache")
        try:
            with open(self.creds_cache_file, 'r') as cfile:
                (self.username, enc_password) = cfile.read().split()
                fernet = Fernet(self.config.get('ndn', 'fernet_key').encode())
                self.password = fernet.decrypt(enc_password.encode()).decode()
                return (self.username, self.password)
        except FileNotFoundError:
            self.log.debug("Failed to find ~/.ndn_creds_cache file")
            self.username = None
            self.password = None
            return (None, None)
        except cryptography.fernet.InvalidToken:
            self.log.warning(
                "Unable to decrypt password from ~/.ndn_creds_cache file.")
            self.log.warning(
                "Please try removing this file and logging in again.")
            return (None, None)

    def save_cached_creds(self, username=None, password=None):
        """
        Cache credentials on the filesystem so they can be reused without bugging the user repeatedly
        """
        self.log.debug("Attempting to save credentials to ~/.ndn_creds_cache")

        fernet = Fernet(self.config.get('ndn', 'fernet_key').encode())
        with open(self.creds_cache_file, 'w') as cfile:
            cfile.write(
                f'{username}\n{fernet.encrypt(password.encode()).decode()}')
        self.username = username
        self.password = password

    def prompt_ldap_login(self, username=None, password=None):
        """
        Request LDAP username and password from the user
        """
        self.username = username or input('LDAP username: ')
        self.password = password or getpass('LDAP password: ')
        # TODO: test ldap login before saving creds, somehow
        # the annoying thing here at the moment is the python-ldap package deps
        self.save_cached_creds(self.username, self.password)

    def vault_login(self, username=None, password=None):
        self.log.debug("Attempting to auth vault client")
        client = hvac.Client(url=self.config.get('ndn', 'vault_url'))
        # try to read token from file first
        if os.path.isfile(self.vault_token_file):
            with open(self.vault_token_file, 'r') as vtoken:
                client.token = vtoken.read()
        try:
            authed = client.is_authenticated()
            self.log.debug(
                "Successfully authed vault client with existing token at %s" % self.vault_token_file)
        except ValueError:
            # if is_authenticated() fails, that probably means the token is bad, so unset it
            client.token = None
            authed = False
            self.log.debug(
                "Failed to auth vault client with existing token at %s" % self.vault_token_file)
        if not authed:
            if username is None:
                username = username or os.environ.get(
                    'VAULT_USERNAME') or input('Vault username: ')
            if password is None:
                password = password or os.environ.get(
                    'VAULT_PASSWORD') or getpass('Vault password: ')
            client.auth.ldap.login(username=username, password=password)
            try:
                self.log.debug(
                    "Writing new vault auth token to %s" % self.vault_token_file)
                with open(self.vault_token_file, 'w') as vtoken:
                    vtoken.write(client.lookup_token()['data']['id'])
            except KeyError:
                pass
        return client

    def api_login(self, username=None, password=None):
        username = username if username else input('API username: ')
        password = password if password else getpass('API password: ')
        self.log.debug("Sending auth login to %s" %
                       self.config.get('ndn', 'auth_api') + '/login')
        auth_resp = self.post_json_url(
            self.config.get('ndn', 'auth_api') + '/login',
            headers={"Content-type": "application/json"},
            json={"username": username, "password": password})
        if auth_resp:
            self.log.debug("API authentication successful!")
            token = auth_resp['access_token']
            self.log.debug("Saving token to %s" % self.jwt_token_file)
            with open(self.jwt_token_file, 'w') as f:
                f.write(token)
            return auth_resp
        else:
            self.log.debug("API authentication API failure: %s" % auth_resp)
            self.log.error(
                "API authentication API failure. Re-run with '-d' for more info.")
            self.log.error(auth_resp.content)
            self.exit_code = 1

    def check_version(self, force=False):
        # try to load upstream version from a cache file, so we don't always have to hit gitlab
        version_cache_file = expanduser('~') + '/.ndn_version'
        upstream_version = {}
        if force:
            self.log.debug(
                'Force requested, skipping version lookup from cache')
        else:
            try:
                with open(version_cache_file, 'r') as vcf:
                    result = json.loads(vcf.read())
                    if int(time.time()) - result['timestamp'] < 300:
                        upstream_version = {
                            'version': result['upstream_version'], 'timestamp': result['timestamp']}
                        self.log.debug(
                            f'Fetched cached upstream version from {version_cache_file}: {upstream_version}')
                    else:
                        raise
            except Exception as err:  # noqa
                self.log.debug(
                    f'Something failed while fetching the cached version: {err}')
                upstream_version = {}

        if not upstream_version:
            self.log.debug('Fetching upstream version from gitlab')
            upstream_version = self.get_url(
                self.config.get('ndn', 'gitlab_api') +
                '/projects/1290/jobs/artifacts/master/raw/version_file?job=version_artifact',
                params={'job': 'version_artifact'},
                headers={'PRIVATE-TOKEN': 'glpat-RW5y_fCHMzy-XweEEnCC'})
            upstream_version['timestamp'] = int(time.time())

        local_version = pkg_resources.get_distribution("ndn").version
        self.log.debug(f'Local version: {local_version}')
        self.log.debug(f'Upstream version: {upstream_version["version"]}')
        if pversion(local_version) < pversion(upstream_version['version']):
            self.log.warning("An update is available for python-ndn-cli!")

        result = {
            'upstream_version': upstream_version['version'],
            'local_version': local_version,
            'timestamp': upstream_version['timestamp']}
        with open(version_cache_file, 'w') as vcf:
            result['timestamp'] = result.get(
                'timestamp', None) or int(time.time())
            vcf.write(json.dumps(result))
        return result


class ndnTest(TestApp, ndn):
    """A sub-class of ndn that is better suited for testing."""

    class Meta:
        label = 'ndn'


def main():
    with ndn() as app:
        # disable log propagation to avoide duplicate log lines
        app.log.backend.propagate = False

        # explicitly load a given plugin for development
        try:
            plugin = app.argv[app.argv.index('--enable-plugin') + 1]
            if plugin in app.plugin.get_loaded_plugins():
                app.log.debug(f'plugin {plugin} already loaded, skipping')
            else:
                app.log.debug(f'loading plugin {plugin}')
                app.plugin.load_plugin(plugin)
        except ValueError:
            # no --enable-plugin passed in, continue
            pass

        if 'auth' not in app.argv and not exists(app.jwt_token_file):
            app.log.warning('auth token file not found at %s' %
                            app.jwt_token_file)
            app.log.warning('you may need to run `ndn login` to get a token')
        if exists(app.jwt_token_file):
            app.log.debug('loading API token from auth file %s' %
                          app.jwt_token_file)
            app.api_token = open(app.jwt_token_file, 'r').read()
            app.AUTH_HEADER = {
                "Content-Type": "application/json",
                'User-agent': 'python-ndn-cli 1.0',
                'Authorization': 'Bearer %s' % app.api_token}
            app.log.debug('loaded API token: %s' % app.api_token)
        app.httpsession = Session()

        # check the version, unless the user explicitly asked to check the version
        # this can be passed as 'ndn version', 'ndn -d version', 'ndn version -f' or similarly
        # as such, we're checking the last 2 args, as this should catch both 'ndn version',
        # and 'ndn -d version -f'
        try:
            if 'version' not in app.argv[-2:]:
                app.check_version()
        # it doesn't matter why this fails, it shouldn't ever interupt the rest of the function
        except:  # noqa
            pass

        # check if we're running on a controller
        # we're using the static auth key, because the user may not have authed
        # if running on a controller
        app.running_on_controller = False
        # app.running_on_controller = gethostname() in app.get_url(
        #    app.config.get('ndn', 'cluster_api') + '/machines/dc-asparagus',
        #    headers={
        #        "Content-Type": "application/json",
        #        'User-agent': 'python-ndn-cli 1.0',
        #        'Authorization': 'Bearer 35JSy5bSUps18Sre8qxUxg8xNkWSTlM'
        #        },
        #    params={'all_machines': True})
        try:
            app.run()

        except AssertionError as e:
            print('AssertionError > %s' % e.args[0])
            app.exit_code = 1

            if app.debug is True:
                import traceback
                traceback.print_exc()

        except ndnError as e:
            print('ndnError > %s' % e.args[0])
            app.exit_code = 1

            if app.debug is True:
                import traceback
                traceback.print_exc()

        except CaughtSignal as e:
            # Default Cement signals are SIGINT and SIGTERM, exit 0 (non-error)
            print('\n%s' % e)
            app.exit_code = 0


if __name__ == '__main__':
    main()
