import os
import sqlalchemy
from sqlalchemy.engine.url import URL
from .vault import vault_login


class NdnSecureDb():
    def __init__(self, name, dbi_string=None, staging=False, read_only=False):
        self.ndnsecure_path = os.environ.get('NDNSECURE_PATH', '/etc/ndn')
        # name of db
        self.name = name
        self.db_name = name
        if staging:
            self.db_name = '%s_staging' % name

        if read_only:
            self.db_name = '%s_clone' % name

        self.ndnsecure_file = os.path.join(
            self.ndnsecure_path, 'db', self.db_name)
        # allow a custom dbi string to be passed instead of using the file
        self.dbi_string = os.getenv('NDN_DBI_%s' % self.name, dbi_string)
        self.db_dict = {
            'user': None,
            'password': None,
            'driver': None,
            'host': None,
            'db': None,
            'port': None
        }

    # returns db creds in dict format

    def get_dict(self):
        # load from env vars in broken out vars
        for k in self.db_dict.keys():
            env_key = "NDN_DB_%s_%s" % (self.db_name, k)
            self.db_dict[k] = os.getenv(
                env_key) if not self.db_dict[k] else self.db_dict[k]

        # if were still missing values fill them in from the db file
        if self.missing_values(self.db_dict):
            try:
                self.db_dict = self.get_from_dbi_string(self.dbi_string)
            except ValueError:
                # failed to fetch creds from env var
                pass

        # if still missing values, fetch from file or vault
        if self.missing_values(self.db_dict):
            if not self.dbi_string:
                if os.environ.get('NDNSECURE_USE_VAULT', False):
                    vclient = vault_login()
                    self.dbi_string = vclient.secrets.kv.v1.read_secret(
                        path='db/%s' % self.db_name, mount_point='secret')['data']['content']
                else:
                    with open(self.ndnsecure_file) as f:
                        self.dbi_string = f.readline().rstrip()
                self.db_dict = self.get_from_dbi_string(self.dbi_string)

        return self.db_dict

    def missing_values(self, db_dict={}):
        optional_values = ['port']
        return [k for k in db_dict.keys() if db_dict[k] is None and k not in optional_values]

    def get_from_dbi_string(self, dbistring):
        try:
            (dbi, user, password) = dbistring.split(" ")
            _, driver, params = dbi.split(':')
            d = dict([p.split("=") for p in params.split(';')])
            d['db'] = d.pop('database', None)
            d['driver'] = driver
            d['user'] = user
            d['password'] = password
            for k in self.db_dict.keys():
                self.db_dict[k] = d[k] if not self.db_dict[k] else self.db_dict[k]

        except Exception:
            pass

        if self.missing_values(self.db_dict):
            raise ValueError(
                'failed to get db credential values for %s' % self.missing_values(self.db_dict))

        return self.db_dict

    # returns db creds in sqlalchemy connection string

    def get_sqlalchemy_conn_uri(self, drivername='mysql+pymysql'):
        db = self.get_dict()
        # we need to give priority to the drivername passed as parameter only
        # if is different from the default
        driver = None
        if drivername != 'mysql+pymysql':
            driver = drivername
        elif db['driver'] == 'mysql':
            driver = 'mysql+pymysql'
        else:
            driver = db['driver']

        if sqlalchemy.__version__ < '1.4.0':
            return URL(
                drivername=driver,
                username=db['user'],
                host=db['host'],
                password=db['password'],
                database=db['db'],
                port=db['port']
            )
        else:

            url = URL.create(
                drivername=driver,
                username=db['user'],
                host=db['host'],
                password=db['password'],
                database=db['db'],
                port=db['port']
            )
            return url.render_as_string(hide_password=False)
