from .core import ServicectlBase, PROD_API_URL
from .schema import JobModel, HistoryModel, SearchJobModel


class Job():
    def __init__(self, job=None, client=None, url=PROD_API_URL):
        job = job or {}
        self.job = JobModel(**job)
        self.client = client or ServicectlBase(url=url)

    def get(self):
        """
        Fetch service_queue_id details from the API
        """

        if not self.job.service_queue_id:
            raise ValueError("Job must have service_queue_id")

        self.job = JobModel(
            **self.client._get(path='/servicectl/jobs/info/%s' % self.job.service_queue_id))
        return self.job

    def schedule(self):
        """
        Schedule a new service_queue job
        """
        self.job = JobModel(**self.client._post(path='/servicectl/jobs/schedule',
                            data=self.job.model_dump(exclude_unset=True)))
        return self.job

    def search(self, search=None):
        """
        Search for a service_queue job(s) based on various parameters
        """
        search = SearchJobModel(**search) if search else self.job

        return [SearchJobModel(**j) for j in self.client._post(path='/servicectl/jobs/search', data=search.model_dump(exclude_unset=True))]

    def delete_cache(self):
        """
        Delete job cache from API
        """
        self.client._delete(path='/servicectl/jobs/info_cache/%s' %
                            self.job.service_queue_id)
        return

    def save(self):
        """
        Save changes to the job object to the API
        """
        self.job = JobModel(**self.client._post(path='/servicectl/jobs/update/%s' %
                            self.job.service_queue_id, data=self.job.model_dump(exclude_unset=True)))

    def update(self):
        """
        Synonym for save()
        """
        self.save()

    def due(self):
        """
        Return a list of jobs due for a given "home:service command"
        """

        if not self.job.home or not self.job.service or not self.job.command:
            raise ValueError("Jobs due requires home, service and command")

        return [JobModel(**j) for j in self.client._get(path='/servicectl/jobs/due', params=self.job.model_dump(exclude_unset=True))]

    def logs(self):
        """
        Fetch logs for a given job
        """

        if not self.job.service_queue_id:
            raise ValueError("logs() function requires service_queue_id")

        return self.client._get(path='/servicectl/jobs/logs/%s' % self.job.service_queue_id)

    def children_finished(self):
        """
        Returns True if all children are finished. Otherwise, returns false.
        """

        if not self.job.service_queue_id:
            raise ValueError(
                "children_finished() function requires service_queue_id")

        return self.client._get(path='/servicectl/jobs/children_finished/%s' % self.job.service_queue_id)

    def children_result(self):
        """
        """

        if not self.job.service_queue_id:
            raise ValueError(
                "children_result() function requires service_queue_id")

        return self.client._get(path='/servicectl/jobs/children_result/%s' % self.job.service_queue_id)

    def release_locks(self):
        """
        Release service_queue_lock on a given service_queue_id
        """

        if not self.job.service_queue_id:
            raise ValueError(
                "release_locks() function requires service_queue_id")

        return self.client._post(path='/servicectl/jobs/release_locks/%s' % self.job.service_queue_id)

    def verify_running(self):
        """
        Verify a given job is running
        """

        if not self.job.service_queue_id:
            raise ValueError(
                "verify_running() function requires service_queue_id")

        return self.client._post(path='/servicectl/jobs/verify_running/%s' % self.job.service_queue_id)

    def died(self):
        """
        Set a job to "died" status
        """

        if not self.job.service_queue_id:
            raise ValueError("died() function requires service_queue_id")

        return self.client._post(path='/servicectl/jobs/died', data=[self.job.service_queue_id])

    def finish(self):
        """
        "Finish" a job, marking the status appropriately
        """

        if not self.job.service_queue_id:
            raise ValueError("finish() function requires service_queue_id")

        return self.client._post(path='/servicectl/jobs/finish', data=[self.job.service_queue_id])

    def fail(self):
        """
        Set a job to "failur" status
        """

        if not self.job.service_queue_id:
            raise ValueError("fail() function requires service_queue_id")

        return self.client._post(path='/servicectl/jobs/fail', data=[self.job.service_queue_id])

    def cancel(self):
        """
        Set a job to "canceled" status
        """

        if not self.job.service_queue_id:
            raise ValueError("cancel() function requires service_queue_id")

        return self.client._post(path='/servicectl/jobs/cancel', data=[self.job.service_queue_id])

    def kill(self):
        """
        "kill" a job, setting status to "killed" and stopping the PID
        """

        if not self.job.service_queue_id:
            raise ValueError("kill() function requires service_queue_id")

        return self.client._post(path='/servicectl/jobs/kill', data=[self.job.service_queue_id])

    def verify_proc_running(self):
        """
        Verify if a process of a running job is still executing
        """

        if not self.job.service_queue_id:
            raise ValueError(
                "verify_proc_running() function requires service_queue_id")

        return self.client._post(path='/servicectl/jobs/verify_proc_running/%s' % self.job.service_queue_id)

    def test_locked(self):
        """
        Test if another process has a lock on a given job
        """

        if not self.job.service_queue_id:
            raise ValueError(
                "test_locked() function requires service_queue_id")

        return self.client._get(path='/servicectl/jobs/test_locked/%s' % self.job.service_queue_id)

    def other_executing(self):
        """
        Test if another process has a lock on a given job
        """

        if not self.job.service_queue_id:
            raise ValueError(
                "other_executing() function requires service_queue_id")

        return self.client._get(path='/servicectl/jobs/other_executing/%s' % self.job.service_queue_id)

    def waiting(self):
        """
        Report if this job is waiting on another job to finish
        """

        if not self.job.service_queue_id:
            raise ValueError("waiting() function requires service_queue_id")

        return self.client._get(path='/servicectl/jobs/waiting/%s' % self.job.service_queue_id)

    def claim(self, proc, fromproc=None):
        """
        Stake a claim for a job
        """

        if not self.job.service_queue_id:
            raise ValueError("claim() function requires service_queue_id")

        return self.client._post(path='/servicectl/jobs/claim/%s' % self.job.service_queue_id, data={'proc': proc, 'fromproc': fromproc})

    def execute(self, sqids=None, controller=None):
        """
        Mark a job as being executed
        """

        if not self.job.service_queue_id and not sqids:
            raise ValueError("execute() function requires service_queue_id")

        # if the user did not pass in a list of sqids, pop in the defined job's sqid
        # note: the API expects sqids to be a list
        if not sqids:
            sqids = [self.job.service_queue_id]
        elif isinstance(sqids, int):
            sqids = [sqids]
        elif not isinstance(sqids, list):
            raise ValueError("sqids must be an int or a list of ints")

        return self.client._post(path='/servicectl/jobs/execute', data=sqids, params={'controller': controller})

    def release(self):
        """
        Release the lock on a job
        """

        if not self.job.service_queue_id:
            raise ValueError("release() function requires service_queue_id")

        return self.client._post(path='/servicectl/jobs/release/%s' % self.job.service_queue_id)

    def active(self, controller=None, channel=None):
        """
        Returns a list of "active" jobs (meaing, currently executing but not finished)
        """

        data = {}
        if controller:
            data['controller'] = controller
        if channel:
            data['channel'] = channel
        result = self.client._get(path='/servicectl/jobs/active', params=data)
        return [Job(job=j, client=self.client) for j in result]

    def claim_lock(self, lockname=None):
        """
        Lock a given job
        """

        if not self.job.service_queue_id:
            raise ValueError("claim_lock() function requires service_queue_id")

        return self.client._post(path='/servicectl/jobs/claim_lock/%s' % self.job.service_queue_id, data={'lockname': lockname})

    def history(self, interval=None, result=None):
        """
        Find the "history" of a given home:service:command, ordered by scheduled
        """

        this_history = HistoryModel(**self.job.model_dump())
        this_history.interval = interval
        this_history.result = result

        final_result = self.client._post(
            path='/servicectl/jobs/history', data=this_history.model_dump(exclude_unset=True))
        return [Job(job=j, client=self.client) for j in final_result]

    def ready_for_pickup(self):
        """
        Returns a list of jobs that are ready to be picked up for processing
        """

        result = self.client._get(path='/servicectl/jobs/ready_for_pickup')
        return [Job(job=j, client=self.client) for j in result]

    def long_running_jobs(self, days=14):
        """
        Returns a list of jobs that have been executing for more than $days
        """

        result = self.client._get(
            path='/servicectl/jobs/long_running', params={'days': days})
        return [Job(job=j, client=self.client) for j in result]
